"""Internal functionality for validating concerns that are common across construct types."""

__all__ = ["validate_common_concerns"]

from viv_compiler import config, errors, external_types, utils
from ._roles import validate_roles
from ._utils import (
    get_all_action_role_names,
    get_all_assigned_scratch_variable_names,
    get_all_negated_expressions,
    get_all_referenced_scratch_variable_names,
    is_non_numeric_expression
)


def validate_common_concerns(
    construct_definition: external_types.ConstructDefinition,
    content_bundle: external_types.ContentBundle
) -> None:
    """Validate the given construct definition along several concerns that are common to all construct types.

    Args:
        construct_definition: The construct definition to validate.
        content_bundle: The compiled content bundle containing the construct definition.

    Returns:
        None. Only returns if no issue was detected downstream.
    """
    # Validate role definitions
    validate_roles(construct_definition=construct_definition)
    # Validate conditions
    _validate_conditions(construct_definition=construct_definition)
    # Validate reactions
    _validate_reactions(construct_definition=construct_definition, content_bundle=content_bundle)
    # Validate action searches
    _validate_action_searches(construct_definition=construct_definition, content_bundle=content_bundle)
    # Validate sifting expressions
    _validate_sifting_expressions(construct_definition=construct_definition, content_bundle=content_bundle)
    # Validate trope-fit expressions
    _validate_trope_fits(construct_definition=construct_definition, content_bundle=content_bundle)
    # Validate application of the group-role decorator
    _validate_group_role_references(construct_definition=construct_definition)
    # Validate reference paths
    _validate_reference_paths(construct_definition=construct_definition)
    # Validate loops
    _validate_loops(construct_definition=construct_definition)
    # Validate assignments
    _validate_assignments(construct_definition=construct_definition)
    # Validate references to scratch variables
    _validate_scratch_variable_usage(construct_definition=construct_definition)
    # Validate negated expressions
    _validate_negated_expressions(construct_definition=construct_definition)
    # Validate chance expressions
    _validate_chance_expressions(construct_definition=construct_definition)


def _validate_conditions(construct_definition: external_types.ConstructDefinition) -> None:
    """Validate the given construct definition's `conditions` field.

    Args:
        construct_definition: A construct definition in a compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The construct definition has an invalid condition.
    """
    # Gather all the conditions, which by now are grouped under roles
    all_conditions: list[external_types.WrappedExpression] = list(
        construct_definition['conditions']['globalConditions']
    )
    for condition_group in construct_definition['conditions']['roleConditions'].values():
        all_conditions += condition_group
    # Validate each one in turn
    for condition in all_conditions:
        for reference in condition['references']:
            role_definition: external_types.RoleDefinition = construct_definition['roles'][reference]
            if role_definition['spawn']:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' has condition that references role '{reference}', "
                    f"which carries the 'spawn' label (prohibited because its entity won't be built until "
                    f"after the conditions have already been evaluated",
                    source=condition['body']['source']
                )


def _validate_reactions(
    construct_definition: external_types.ConstructDefinition,
    content_bundle: external_types.ContentBundle
) -> None:
    """Validate the given construct definition's use of reactions.

    Args:
        construct_definition: A construct definition in a compiled content bundle.
        content_bundle: A compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The construct definition has an invalid reaction.
        VivCompileError: The construct definition has an invalid reaction.
    """
    # Collect all reactions
    all_reaction_expressions: list[external_types.Reaction] = utils.get_all_full_expressions_of_type(
        expression_type=external_types.ExpressionDiscriminator.REACTION,
        ast_chunk=construct_definition
    )
    # If this is not a construct type allowing reactions, ensure that there are no reaction declarations whatsoever
    if all_reaction_expressions:
        if construct_definition['type'] not in config.CONSTRUCT_TYPES_PERMITTING_REACTIONS:
            raise errors.VivCompileError(
                f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                f"'{construct_definition['name']}' contains a reaction declaration (not permitted in "
                f"this construct type)",
                source=all_reaction_expressions[0]['source']
            )
    # Make sure that all reactions are housed in the proper fields
    if construct_definition['type'] == external_types.ConstructDiscriminator.ACTION:
        for field_name, field_value in construct_definition.items():
            if field_name not in config.ACTION_FIELDS_PERMITTING_REACTIONS:
                reactions_in_this_field: list[external_types.ReactionValue] = utils.get_all_expressions_of_type(
                    expression_type=external_types.ExpressionDiscriminator.REACTION,
                    ast_chunk=field_value
                )
                if reactions_in_this_field:
                    raise errors.VivCompileError(
                        f"Action '{construct_definition['name']}' has reaction in '{field_name}' field "
                        f"(only allowed in 'reactions' field)"
                    )
    # Validate each reaction in turn
    for reaction_expression in all_reaction_expressions:
        reaction: external_types.ReactionValue = reaction_expression['value']
        reaction_source = reaction_expression['source']
        # Make sure the reaction queues a valid construct
        queued_construct_name = reaction['targetName']
        if reaction['targetType'] == external_types.ConstructDiscriminator.ACTION:
            try:
                queued_construct_definition = content_bundle["actions"][queued_construct_name]
            except KeyError:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' has reaction that queues undefined action: "
                    f"'{queued_construct_name}'",
                    source=reaction_source
                )
        elif reaction['targetType'] == external_types.ConstructDiscriminator.PLAN:
            try:
                queued_construct_definition = content_bundle["plans"][queued_construct_name]
            except KeyError:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' has reaction that queues undefined plan: "
                    f"'{queued_construct_name}'",
                    source=reaction_source
                )
        elif reaction['targetType'] == external_types.ConstructDiscriminator.ACTION_SELECTOR:
            try:
                queued_construct_definition = content_bundle["actionSelectors"][queued_construct_name]
            except KeyError:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' has reaction that queues undefined action "
                    f"selector: '{queued_construct_name}'",
                    source=reaction_source
                )
        elif reaction['targetType'] == external_types.ConstructDiscriminator.PLAN_SELECTOR:
            try:
                queued_construct_definition = content_bundle["planSelectors"][queued_construct_name]
            except KeyError:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' has reaction that queues undefined plan selector:"
                    f" '{queued_construct_name}'",
                    source=reaction_source
                )
        else:
            raise errors.VivCompileError(
                f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                f"'{construct_definition['name']}' has reaction that queues construct with unexpected "
                f"type: '{reaction['targetType']}'",
                source=reaction_source
            )
        # If applicable, make sure the reaction binds an initiator role
        if queued_construct_definition['type'] in config.CONSTRUCT_TYPES_USING_INITIATOR_ROLES:
            if queued_construct_definition['initiator'] not in reaction['bindings']['roles']:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' has '{queued_construct_name}' reaction that fails to precast "
                    f"its initiator role '{queued_construct_definition['initiator']}' (required when queueing a "
                    f"{utils.CONSTRUCT_LABEL[queued_construct_definition['type']].lower()})",
                    source=reaction_source
                )
        # Detect missing required roles
        if not reaction['bindings']['partial']:
            for role_definition in queued_construct_definition["roles"].values():
                if role_definition['min'] > 0 and role_definition['name'] not in reaction['bindings']['roles']:
                    raise errors.VivCompileError(
                        f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                        f"'{construct_definition['name']}' has '{queued_construct_name}' reaction "
                        f"that fails to cast required role '{role_definition['name']}' "
                        f"(use 'partial' to pass a partial cast)",
                        source=reaction_source
                    )
        # Detect missing 'precast' roles
        for role_definition in queued_construct_definition['roles'].values():
            if role_definition['precast']:
                if role_definition['name'] not in reaction['bindings']['roles']:
                    raise errors.VivCompileError(
                        f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                        f"'{construct_definition['name']}' has '{queued_construct_name}' reaction "
                        f"that fails to precast one of its 'precast' roles: '{role_definition['name']}'",
                        source=reaction_source
                    )
        # Detect extra roles
        for role_name in reaction['bindings']['roles']:
            if role_name not in queued_construct_definition['roles']:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' has '{queued_construct_name}' reaction that precasts "
                    f"role '{role_name}', but the target construct has no role by that name",
                    source=reaction_source
                )
        # Validate reaction priority
        if reaction['priority']:
            if reaction['targetType'] not in (
                external_types.ConstructDiscriminator.ACTION,
                external_types.ConstructDiscriminator.ACTION_SELECTOR
            ):
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' has '{queued_construct_name}' reaction with 'priority' "
                    f"field (only allowed for reactions that queue actions or action selectors)",
                    source=reaction_source
                )
            if is_non_numeric_expression(expression=reaction['priority']):
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' has '{queued_construct_name}' reaction with invalid expression "
                    f"type for 'priority' field (value will be non-numeric)",
                    source=reaction_source
                )
        # Validate the optional 'time' field, if it's present
        time_of_day_constraints = 0
        time_frame_from_action_constraints = 0
        time_frame_from_hearing_constraints = 0
        if reaction['time']:
            if reaction['targetType'] not in (
                external_types.ConstructDiscriminator.ACTION,
                external_types.ConstructDiscriminator.ACTION_SELECTOR
            ):
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' has '{queued_construct_name}' reaction with 'time' field "
                    f"(only allowed for reactions that queue actions or action selectors)",
                    source=reaction_source
                )
            for temporal_constraint in reaction['time']:
                if temporal_constraint['type'] == external_types.TemporalStatementDiscriminator.TIME_OF_DAY:
                    time_of_day_constraints += 1
                    if time_of_day_constraints > 1:
                        raise errors.VivCompileError(
                            f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                            f"'{construct_definition['name']}' has '{queued_construct_name}' reaction "
                            f"with multiple time-of-day temporal constraints (only one is allowed)",
                            source=reaction_source
                        )
                elif temporal_constraint['useActionTimestamp']:
                    time_frame_from_action_constraints += 1
                    if time_frame_from_action_constraints > 1:
                        raise errors.VivCompileError(
                            f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                            f"'{construct_definition['name']}' has '{queued_construct_name}' reaction "
                            f"with multiple 'from action' temporal constraints (only one is allowed)",
                            source=reaction_source
                        )
                else:  # temporal_constraint['useActionTimestamp'] == false
                    time_frame_from_hearing_constraints += 1
                    if time_frame_from_hearing_constraints > 1:
                        raise errors.VivCompileError(
                            f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                            f"'{construct_definition['name']}' has '{queued_construct_name}' reaction "
                            f"with multiple 'from hearing' temporal constraints (only one is allowed)",
                            source=reaction_source
                        )
        # Validate the optional 'location' field, if it's present
        if reaction['location']:
            if reaction['targetType'] not in (
                external_types.ConstructDiscriminator.ACTION,
                external_types.ConstructDiscriminator.ACTION_SELECTOR
            ):
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' has '{queued_construct_name}' reaction with 'location' "
                    f"field (only allowed for reactions that queue actions or action selectors)",
                    source=reaction_source
                )
            for set_predicate in reaction['location']:
                if set_predicate["operator"] == "all":
                    raise errors.VivCompileError(
                        f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                        f"'{construct_definition['name']}' has '{queued_construct_name}' reaction that uses the"
                        f"'all' operator, but this is not allowed because this field corresponds to a singular value "
                        f"(use 'exactly' to specify a single possibility)",
                        source=reaction_source
                    )
                if set_predicate['operator'] == "exactly":
                    if len(set_predicate['operand']) > 1:
                        raise errors.VivCompileError(
                            f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                            f"'{construct_definition['name']}' has '{queued_construct_name}' reaction that uses the"
                            f"'exactly' operator in 'location' field, but then has multiple operand components "
                            f"(only a single location can be specified when 'exactly' is used)",
                            source=reaction_source
                        )
                    if len(reaction['location']) > 1:
                        raise errors.VivCompileError(
                            f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                            f"'{construct_definition['name']}' has '{queued_construct_name}' reaction that uses the"
                            f"'exactly' operator in 'location' field, but then has additional criteria there "
                            f"('exactly' must be alone when present)",
                            source=reaction_source
                        )
                for operand_component in set_predicate['operand']:
                    if set_predicate['operand'].count(operand_component) > 1:
                        raise errors.VivCompileError(
                            f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                            f"'{construct_definition['name']}' has '{queued_construct_name}' reaction with duplicate "
                            f"value following '{set_predicate['operator']}' operator in the 'location' field",
                            source=reaction_source
                        )
        # Validate the optional 'repeatLogic' field, if it's present
        if reaction['repeatLogic']:
            if not reaction['repeatLogic']["conditions"]:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' has '{queued_construct_name}' reaction with 'repeat' "
                    f"field that is missing required 'if' subfield",
                    source=reaction_source
                )
            if not reaction['repeatLogic']["maxInstances"]:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' has '{queued_construct_name}' reaction with 'repeat' "
                    f"field that is missing required 'max' subfield",
                    source=reaction_source
                )


def _validate_action_searches(
    construct_definition: external_types.ConstructDefinition,
    content_bundle: external_types.ContentBundle
) -> None:
    """Validate the given construct definition's usage of action searches.

    Args:
        construct_definition: A construct definition in a compiled content bundle.
        content_bundle: A compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The construct definition has an invalid action search.
        VivCompileError: The construct definition has an invalid action search
    """
    all_action_search_full_expressions: list[external_types.ActionSearch] = utils.get_all_full_expressions_of_type(
        expression_type=external_types.ExpressionDiscriminator.ACTION_SEARCH,
        ast_chunk=construct_definition
    )
    for action_search_full_expression in all_action_search_full_expressions:
        action_search_expression: external_types.ActionSearchValue = action_search_full_expression['value']
        action_search_source = action_search_full_expression['source']
        # Validate the search-domain preparation policy. This must come first, because it applies
        # even if there is no target query, which causes an immediate return below.
        _validate_search_domain_declaration(
            construct_definition=construct_definition,
            search_domain_declaration=action_search_expression['searchDomain'],
            in_sifting=False,
            source=action_search_source
        )
        # If this is a search without a query, move on
        query_name = action_search_expression['queryName']
        if not query_name:
            continue
        # Otherwise, check if it targets an undefined query
        if query_name and query_name not in content_bundle["queries"]:
            raise errors.VivCompileError(
                f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                f"'{construct_definition['name']}' references undefined query: '{query_name}'",
                source=action_search_source
            )
        # Detect missing required roles
        query_definition = content_bundle["queries"][query_name]
        if not action_search_expression['bindings']['partial']:
            for role_definition in query_definition["roles"].values():
                if role_definition['min'] > 0:
                    if role_definition['name'] not in action_search_expression['bindings']['roles']:
                        raise errors.VivCompileError(
                            f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                            f"'{construct_definition['name']}' fails to cast required role '{role_definition['name']}' "
                            f"when running search query '{query_name}' (use 'partial' to pass a partial cast)",
                            source=action_search_source
                        )
        # Detect missing 'precast' roles
        for role_definition in query_definition['roles'].values():
            if role_definition['precast']:
                if role_definition['name'] not in action_search_expression['bindings']['roles']:
                    raise errors.VivCompileError(
                        f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                        f"'{construct_definition['name']}' runs search query '{query_name}', "
                        f"but fails to precast one of its 'precast' roles: '{role_definition['name']}'",
                        source=action_search_source
                    )
        # Detect extra roles
        for role_name in action_search_expression['bindings']['roles']:
            if role_name not in query_definition['roles']:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' precasts role '{role_name}' when running search "
                    f"query '{query_name}', but the query has no role by that name",
                    source=action_search_source
                )
        # If the query uses memory fields, ensure that the search domain is a character's memories
        if query_definition['salience'] or query_definition['associations']:
            if not action_search_expression['searchDomain']['expression']:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' attempts to run search query '{query_name}', which uses the "
                    f"'salience' and/or 'associations' memory fields, but the action search does not specify a "
                    f"character whose memories will serve as the search domain (use the 'over' field to specify this)",
                    source=action_search_source
                )


def _validate_sifting_expressions(
    construct_definition: external_types.ConstructDefinition,
    content_bundle: external_types.ContentBundle
) -> None:
    """Validate the given construct definition's usage of sifting expressions.

    Args:
        construct_definition: A construct definition in a compiled content bundle.
        content_bundle: A compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The construct definition has an invalid sifting expression.
        VivCompileError: The construct definition has an invalid sifting expression.
    """
    all_sifting_full_expressions: list[external_types.Sifting] = utils.get_all_full_expressions_of_type(
        expression_type=external_types.ExpressionDiscriminator.SIFTING,
        ast_chunk=construct_definition
    )
    for sifting_full_expression in all_sifting_full_expressions:
        sifting_expression: external_types.SiftingValue = sifting_full_expression['value']
        sifting_source = sifting_full_expression['source']
        # Retrieve the name of the sifting pattern referenced in the expression
        pattern_name = sifting_expression['patternName']
        # Check if targets to undefined sifting pattern
        if pattern_name not in content_bundle["siftingPatterns"]:
            raise errors.VivCompileError(
                f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                f"'{construct_definition['name']}' references undefined sifting pattern: '{pattern_name}'",
                source=sifting_source
            )
        # Validate the search-domain preparation policy
        _validate_search_domain_declaration(
            construct_definition=construct_definition,
            search_domain_declaration=sifting_expression['searchDomain'],
            in_sifting=True,
            source=sifting_source
        )
        # Detect missing required roles
        pattern_definition = content_bundle["siftingPatterns"][pattern_name]
        if not sifting_expression['bindings']['partial']:
            for role_definition in pattern_definition["roles"].values():
                if role_definition['min'] > 0:
                    if role_definition['name'] not in sifting_expression['bindings']['roles']:
                        raise errors.VivCompileError(
                            f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                            f"'{construct_definition['name']}' fails to cast required role '{role_definition['name']}' "
                            f"when attempting to sift pattern '{pattern_name}' "
                            f"(use 'partial' to pass a partial cast)",
                            source=sifting_source
                        )
        # Detect missing precast roles
        for role_definition in pattern_definition['roles'].values():
            if role_definition['precast']:
                if role_definition['name'] not in sifting_expression['bindings']['roles']:
                    raise errors.VivCompileError(
                        f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                        f"'{construct_definition['name']}' attempts to sift pattern '{pattern_name}', "
                        f"but fails to precast one of its 'precast' roles: '{role_definition['name']}'",
                        source=sifting_source
                    )
        # Detect extra roles
        for role_name in sifting_expression['bindings']['roles']:
            if role_name not in pattern_definition['roles']:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' precasts role '{role_name}' when attempting to sift pattern "
                    f"'{pattern_name}', but the pattern has no role by that name",
                    source=sifting_source
                )


def _validate_trope_fits(
    construct_definition: external_types.ConstructDefinition,
    content_bundle: external_types.ContentBundle
) -> None:
    """Validate the given construct definition's usage of trope-fit expressions.

    Note: In the course of resolving any positional bindings, we will have already flagged
    any cases of trope fits that target undefined tropes.

    Args:
        construct_definition: A construct definition in a compiled content bundle.
        content_bundle: A compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The construct definition has an invalid trope-fit expression.
        VivCompileError: The construct definition has an invalid trope-fit expression.
    """
    all_trope_fit_full_expressions: list[external_types.TropeFit] = utils.get_all_full_expressions_of_type(
        expression_type=external_types.ExpressionDiscriminator.TROPE_FIT,
        ast_chunk=construct_definition
    )
    for trope_fit_full_expression in all_trope_fit_full_expressions:
        trope_fit_expression: external_types.TropeFitValue = trope_fit_full_expression['value']
        trope_fit_source = trope_fit_full_expression['source']
        # Retrieve the definition of the target trope
        trope_name = trope_fit_expression['tropeName']
        trope_definition = content_bundle["tropes"][trope_name]
        # Detect missing required roles
        if not trope_fit_expression['bindings']['partial']:
            for role_definition in trope_definition["roles"].values():
                if role_definition['min'] > 0:
                    if role_definition['name'] not in trope_fit_expression['bindings']['roles']:
                        raise errors.VivCompileError(
                            f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                            f"'{construct_definition['name']}' fails to cast required role '{role_definition['name']}' "
                            f"when attempting to fit trope '{trope_name}' (use 'partial' to pass a partial cast)",
                            source=trope_fit_source
                        )
        # Detect missing precast roles
        for role_definition in trope_definition['roles'].values():
            if role_definition['precast']:
                if role_definition['name'] not in trope_fit_expression['bindings']['roles']:
                    raise errors.VivCompileError(
                        f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                        f"'{construct_definition['name']}' attempts to fit trope '{trope_name}', "
                        f"but fails to precast one of its 'precast' roles: '{role_definition['name']}'",
                        source=trope_fit_source
                    )
        # Detect extra roles
        for role_name in trope_fit_expression['bindings']['roles']:
            if role_name not in trope_definition['roles']:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' precasts role '{role_name}' when attempting to fit trope "
                    f"'{trope_name}', but the trope has no role by that name",
                    source=trope_fit_source
                )


def _validate_group_role_references(construct_definition: external_types.ConstructDefinition) -> None:
    """Validate the given construct definition's usage of the group-role decorator (or lack thereof).

    Args:
        construct_definition: A construct definition in a compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The construct definition does not properly apply the group-role decorator.
    """
    # Retrieve all references to group roles
    all_references: list[external_types.ReferenceValue] = utils.get_all_expressions_of_type(
        expression_type=external_types.ExpressionDiscriminator.ENTITY_REFERENCE,
        ast_chunk=construct_definition
    )
    all_references += utils.get_all_expressions_of_type(
        expression_type=external_types.ExpressionDiscriminator.SYMBOL_REFERENCE,
        ast_chunk=construct_definition
    )
    all_group_role_names = [reference['anchor'] for reference in all_references if reference['group']]
    # Make sure the group-role decorator is only used on roles that can cast multiple entities. Here, we'll also
    # include all our special roles, since these are always singletons.
    for group_role_name in all_group_role_names:
        error_message = (
            f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
            f"uses the group-role decorator '*' on a singleton role (one with a max of 1): '{group_role_name}'"
        )
        if group_role_name in config.SPECIAL_ROLE_NAMES:
            raise errors.VivCompileError(error_message)
        for role_definition in construct_definition['roles'].values():
            if role_definition['name'] == group_role_name:
                if role_definition['max'] < 2:
                    raise errors.VivCompileError(error_message)
                break
    # Make sure that all other references to group roles *do* use the group-role decorator
    all_singleton_role_names = [reference['anchor'] for reference in all_references if not reference['group']]
    for singleton_role_name in all_singleton_role_names:
        for role_definition in construct_definition['roles'].values():
            if role_definition['name'] == singleton_role_name:
                if role_definition['max'] > 1:
                    raise errors.VivCompileError(
                        f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                        f"'{construct_definition['name']} references a group role without using the group-role "
                        f"decorator '*': '{singleton_role_name}'"
                    )
                break


def _validate_reference_paths(construct_definition: external_types.ConstructDefinition) -> None:
    """Validate the given construct definition's usage of reference paths.

    Args:
        construct_definition: A construct definition in a compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: A reference has a pointer as its first path component.
    """
    # Collect all entity and symbol references
    all_references: list[external_types.ReferenceValue] = utils.get_all_expressions_of_type(
        expression_type=external_types.ExpressionDiscriminator.ENTITY_REFERENCE,
        ast_chunk=construct_definition
    )
    all_references += utils.get_all_expressions_of_type(
        expression_type=external_types.ExpressionDiscriminator.SYMBOL_REFERENCE,
        ast_chunk=construct_definition
    )
    # Make sure no reference has a pointer as its first path component. While a syntax like `@person->boss`
    # might seem required, because `@person` evaluates to an entity ID, Viv enables authors to directly
    # access the properties of the anchor entity, as in `@person.boss`. To reduce confusion, Viv enforces
    # that this pattern be used exclusively. Under the hood, this is supported by the interpreter hydrating
    # the anchor entity ID prior to walking a reference path.
    for reference in all_references:
        if not reference['path']:
            continue
        pointer_type = external_types.ReferencePathComponentDiscriminator.REFERENCE_PATH_COMPONENT_POINTER
        if reference['path'][0]['type'] == pointer_type:
            raise errors.VivCompileError(
                f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
                f"has reference '@{reference['anchor']}' with a pointer as its first path component "
                f"(use '@{reference['anchor']}.{reference['path'][0]['propertyName']}' instead)"
            )


def _validate_loops(construct_definition: external_types.ConstructDefinition) -> None:
    """Validate the given construct definition's usage of loops.

    Note: The grammar already enforces that loop bodies may not be empty.

    Args:
        construct_definition: A construct definition from a compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The construct definition has an invalid loop.
    """
    all_loop_expressions: list[external_types.Loop] = utils.get_all_full_expressions_of_type(
        expression_type=external_types.ExpressionDiscriminator.LOOP,
        ast_chunk=construct_definition
    )
    for loop_expression in all_loop_expressions:
        loop: external_types.LoopValue = loop_expression['value']
        loop_source = loop_expression['source']
        # Detect attempts to loop over single-entity role references, i.e., ones using `@role` notation
        # without a group-role decorator (`@role*`) or a path (`@role.path.to.collection`). Of course,
        # the path could point to a non-iterable value, but that's a runtime issue that depends on the
        # domain of the host application, so there's nothing we can do to detect that kind of issue here.
        if loop['iterable']['type'] == external_types.ExpressionDiscriminator.ENTITY_REFERENCE:
            if not loop['iterable']['value']['group'] and not loop['iterable']['value']['path']:
                role_name = loop['iterable']['value']['anchor']
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' attempts to loop over a non-unpacked role: '{role_name}' "
                    f"(perhaps use * here)",
                    source=loop_source
                )
        # Detect cases of a loop variable shadowing the name of a role from the same action
        if loop['variable']['name'] in get_all_action_role_names(action_definition=construct_definition):
            raise errors.VivCompileError(
                f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                f"'{construct_definition['name']}' has loop with variable name that shadows role name "
                f"'{loop['variable']['name']}' (this is prohibited)",
                source=loop_source
            )


def _validate_assignments(construct_definition: external_types.ConstructDefinition) -> None:
    """Validate the given construct definition's usage of assignments.

    Args:
        construct_definition: A construct definition in a compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The construct definition has an invalid assignment.
    """
    # Collect all assignments
    all_assignment_expressions: list[external_types.Assignment] = utils.get_all_full_expressions_of_type(
        expression_type=external_types.ExpressionDiscriminator.ASSIGNMENT,
        ast_chunk=construct_definition
    )
    # If this is not an action, ensure that there are no assignments whatsoever
    if construct_definition["type"] != external_types.ConstructDiscriminator.ACTION:
        if all_assignment_expressions:
            raise errors.VivCompileError(
                f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                f"'{construct_definition['name']}' contains an assignment (only permitted in actions)",
                source=all_assignment_expressions[0]['source']
            )
        return
    # Make sure that all assignments are housed in the proper fields
    for field_name, field_value in construct_definition.items():
        if field_name not in config.ACTION_FIELDS_PERMITTING_ASSIGNMENTS:
            assignments_in_field: list[external_types.AssignmentValue] = utils.get_all_expressions_of_type(
                expression_type=external_types.ExpressionDiscriminator.ASSIGNMENT,
                ast_chunk=field_value
            )
            if assignments_in_field:
                raise errors.VivCompileError(
                    f"Action '{construct_definition['name']}' has assignment in '{field_name}' field "
                    f"(only allowed in {', '.join(config.ACTION_FIELDS_PERMITTING_ASSIGNMENTS)})"
                )
    # Validate each assignment in turn
    for assignment_expression in all_assignment_expressions:
        assignment: external_types.AssignmentValue = assignment_expression['value']
        assignment_source = assignment_expression['source']
        assignment_lhs: external_types.ReferenceValue = assignment['left']['value']
        anchor = assignment_lhs['anchor']
        path = assignment_lhs['path']
        # Detect attempts to set a local variable. Note that we do allow assignments that are anchored
        # in local variables and also have a path, since this enables the common pattern of looping
        # over a group role to execute effects for each entity cast in the group role.
        if assignment['left']['value']['local'] and not path:
            raise errors.VivCompileError(
                f"Assignment expression in action '{construct_definition['name']}' sets local variable '{anchor}', "
                f"but local variables can only be set in loops, 'saliences' headers, and 'associations' headers",
                source=assignment_source
            )
        if anchor != config.ACTION_SELF_REFERENCE_ROLE_NAME and anchor in construct_definition['roles']:
            # Detect attempts to recast a role via assignment (not allowed)
            if not path:
                raise errors.VivCompileError(
                    f"Assignment expression in action '{construct_definition['name']}' recasts "
                    f"role '{anchor}' (this is prohibited)",
                    source=assignment_source
                )
            # Detect attempts to set data on a complex symbol role (currently prohibited)
            anchor_role_definition = construct_definition['roles'][anchor]
            if anchor_role_definition['entityType'] == external_types.RoleEntityType.SYMBOL:
                raise errors.VivCompileError(
                    f"Assignment expression in action '{construct_definition['name']}' has "
                    f"symbol role on its left-hand side (this is currently prohibited): '{anchor}'",
                    source=assignment_source
                )
        # Detect a trailing eval fail-safe marker, which is bizarre and probably an authoring error
        if path and path[-1]['failSafe']:
            raise errors.VivCompileError(
                f"LHS of assignment expression in action '{construct_definition['name']}' has "
                f"a trailing eval fail-safe marker '?' (only allowed within, and not following, the LHS)",
                source=assignment_source
            )


def _validate_scratch_variable_usage(construct_definition: external_types.ConstructDefinition) -> None:
    """Validate the given construct definition's usage of scratch variables.

    This procedure ensures that any scratch variable that is referenced is assigned *somewhere* in
    the construct definition, but it does not ensure that the variable is assigned before it's used.

    Args:
        construct_definition: A construct definition in a compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The construct definition makes invalid use of scratch variables.
    """
    # Retrieve the names of all scratch variables that are set anywhere in the construct definition
    all_assigned_scratch_variable_names = set(get_all_assigned_scratch_variable_names(ast_chunk=construct_definition))
    # Retrieve the names of all scratch variables that are referenced anywhere in the action definition
    all_referenced_scratch_variable_names = set(get_all_referenced_scratch_variable_names(
        ast_chunk=construct_definition
    ))
    # If this is not an action, ensure that there is no usage of scratch variables whatsoever
    if construct_definition["type"] != external_types.ConstructDiscriminator.ACTION:
        if all_assigned_scratch_variable_names or all_referenced_scratch_variable_names:
            raise errors.VivCompileError(
                f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                f"'{construct_definition['name']}' makes use of scratch variables (only permitted in actions)"
            )
        return
    # Flag any cases of a scratch variable being referenced without being assigned anywhere
    referenced_but_not_assigned = all_referenced_scratch_variable_names - all_assigned_scratch_variable_names
    if referenced_but_not_assigned:
        snippet = "'" + "', '".join(sorted(referenced_but_not_assigned)) + "'"
        raise errors.VivCompileError(
            f"Action '{construct_definition['name']}' references the following scratch variables without ever "
            f"assigning them: {snippet}"
        )


def _validate_search_domain_declaration(
    construct_definition: external_types.ConstructDefinition,
    search_domain_declaration: external_types.SearchDomainDeclaration,
    in_sifting: bool,
    source: external_types.SourceAnnotations | None = None
) -> None:
    """Validates the given search-domain declaration.

    Args:
        construct_definition: Definition for the construct containing the search-domain declaration.
        search_domain_declaration: The search-domain declaration to validate.
        in_sifting: Whether the search is part of a sifting, as opposed to an action search.
        source: Source annotations for the expression containing the search-domain declaration.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: An unexpected construct type is encountered.
        VivCompileError: The construct definition has an invalid search-domain declaration.
    """
    match construct_definition['type']:
        case external_types.ConstructDiscriminator.ACTION:
            prohibited_policy = external_types.SearchDomainPreparationPolicy.INHERIT
        case external_types.ConstructDiscriminator.ACTION_SELECTOR:
            prohibited_policy = external_types.SearchDomainPreparationPolicy.INHERIT
        case external_types.ConstructDiscriminator.PLAN:
            prohibited_policy = external_types.SearchDomainPreparationPolicy.INHERIT
        case external_types.ConstructDiscriminator.PLAN_SELECTOR:
            prohibited_policy = external_types.SearchDomainPreparationPolicy.INHERIT
        case external_types.ConstructDiscriminator.QUERY:
            prohibited_policy = external_types.SearchDomainPreparationPolicy.CHRONICLE
        case external_types.ConstructDiscriminator.SIFTING_PATTERN:
            prohibited_policy = external_types.SearchDomainPreparationPolicy.CHRONICLE
        case external_types.ConstructDiscriminator.TROPE:
            # A trope works with any policy, though there is the chance of a discrepancy at runtime,
            # if a trope fit in a complex expression chain and does not specify the right search-domain
            # preparation policy. This will be flagged if it occurs.
            return
        case _:
            raise errors.VivCompileError(
                f"Unexpected construct type: {construct_definition['type']}",
                source=source
            )
    if search_domain_declaration['policy'] == prohibited_policy:
        context_snippet = "action search" if in_sifting else "sifting"
        raise errors.VivCompileError(
            f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
            f"uses 'over: {prohibited_policy.upper()}' in a {context_snippet} (prohibited for this construct type)",
            source=source
        )


def _validate_negated_expressions(construct_definition: external_types.ConstructDefinition) -> None:
    """Validate the given construct definition's usage of expression negation.

    Args:
        construct_definition: A construct definition in a compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The construct definition negates an expression that cannot be negated.
    """
    for negated_expression in get_all_negated_expressions(ast_chunk=construct_definition):
        if negated_expression['type'] not in config.NEGATABLE_EXPRESSION_TYPES:
            raise errors.VivCompileError(
                f"Expression of type '{negated_expression['type']}' in "
                f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                f"'{construct_definition['name']}' is negated, but this is not allowed (only the following "
                f"expression types support negation: {', '.join(config.NEGATABLE_EXPRESSION_TYPES)}):\n\n"
                f"{negated_expression}",
                source=negated_expression.get('source')
            )


def _validate_chance_expressions(construct_definition: external_types.ConstructDefinition) -> None:
    """Validate the given construct definition's usage of chance expressions.

    Args:
        construct_definition: A construct definition in a compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The construct definition has an invalid chance expression.
    """
    all_chance_expressions: list[external_types.ChanceExpression] = utils.get_all_full_expressions_of_type(
        expression_type=external_types.ExpressionDiscriminator.CHANCE_EXPRESSION,
        ast_chunk=construct_definition
    )
    for chance_expression in all_chance_expressions:
        chance_value: float = chance_expression['value']
        chance_source = chance_expression['source']
        if chance_value < 0.0 or chance_value > 1.0:
            raise errors.VivCompileError(
                f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                f"'{construct_definition['name']}' has chance expression with chance value outside of "
                f"the range [0, 100]: '{chance_value * 100}%'",
                source=chance_source
            )



"""Internal functionality for validating trope definitions."""

__all__ = ["validate_trope_definitions"]

from viv_compiler import external_types
from ._common import validate_common_concerns


def validate_trope_definitions(content_bundle: external_types.ContentBundle) -> None:
    """Validate the trope definitions in the given compiled content bundle.

    Args:
        content_bundle: A compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.
    """
    for trope_definition in content_bundle["tropes"].values():
        # Validate general issues that pertain to all construct types
        validate_common_concerns(construct_definition=trope_definition, content_bundle=content_bundle)



"""Handles inheritance between action definitions."""

__all__ = ["handle_inheritance"]

import copy
from typing import Any, Final

from viv_compiler import errors, external_types, internal_types, sentinels, validation


def handle_inheritance(combined_ast: internal_types.CombinedAST) -> None:
    """Modify the given combined AST in place to honor all inheritance declarations.

    Any template actions (definitions prepended with the `template` marker) will be deleted
    as a final step in this procedure.

    Note: Currently only action definitions support inheritance.

    Args:
        combined_ast: An abstract syntax tree produced by the Visitor class, integrating the
            respective ASTs of any included files.

    Returns:
        None (construct definitions are modified in place).
    """
    validation.prevalidate_action_inheritance(combined_ast=combined_ast)
    intermediate_action_definitions = combined_ast["actions"]
    outstanding_child_action_names = [action['name'] for action in combined_ast["actions"] if action['parent']]
    while outstanding_child_action_names:
        made_progress = False
        for i, child_action_definition in enumerate(intermediate_action_definitions):
            child_action_definition: internal_types.IntermediateChildActionDefinition
            if child_action_definition['name'] not in outstanding_child_action_names:
                continue
            if child_action_definition['parent'] in outstanding_child_action_names:
                # This action's parent itself has a parent, and so we need to wait for the parent
                # to inherit its material first (and so on if there's further dependencies).
                continue
            try:
                parent_action_definition = next(
                    action for action in intermediate_action_definitions
                    if action['name'] == child_action_definition['parent']
                )
            except StopIteration:
                raise errors.VivCompileError(
                    f"Action '{child_action_definition['name']}' declares undefined parent action "
                    f"'{child_action_definition['parent']}'"
                )
            # Handle any join directives
            merged_action_definition = _merge_action_definitions(
                child_action_definition=child_action_definition,
                parent_action_definition=parent_action_definition
            )
            # Handle any role-renaming declarations
            _handle_role_renaming_declarations(
                merged_action_definition=merged_action_definition,
                parent_action_name=parent_action_definition['name'],
                parent_role_names=set(parent_action_definition['roles'].keys())
            )
            # Save the finalized child action
            intermediate_action_definitions[i] = merged_action_definition
            outstanding_child_action_names.remove(child_action_definition['name'])
            made_progress = True
        # If we made no progress this pass, the remaining names participate in a cycle.
        if not made_progress:
            cycle_actions = "'" + ", ".join(outstanding_child_action_names) + "'"
            raise errors.VivCompileError(f"Inheritance cycle detected among actions: {cycle_actions}")
    # Finally, delete any template actions
    combined_ast["actions"] = [action for action in combined_ast["actions"] if not action['_template']]


def _merge_action_definitions(
    child_action_definition: internal_types.IntermediateChildActionDefinition,
    parent_action_definition: internal_types.IntermediateActionDefinition
) -> internal_types.IntermediateActionDefinition:
    """Clone the given action definitions and return a merged one that honors the author's inheritance declarations.

    Args:
        child_action_definition: The child action definition that will inherit from the given parent definition.
        parent_action_definition: The parent action definition from which the given child definition will inherit.

    Returns:
        A merged action definition that honors the author's inheritance declarations.
    """
    field_name_to_join_directive_sentinel: Final = {
        "tags": sentinels.CHILD_JOIN_DIRECTIVE_TAGS,
        "roles": sentinels.CHILD_JOIN_DIRECTIVE_ROLES,
        "_conditions_raw": sentinels.CHILD_JOIN_DIRECTIVE_CONDITIONS,
        "scratch": sentinels.CHILD_JOIN_DIRECTIVE_SCRATCH,
        "_effects_raw": sentinels.CHILD_JOIN_DIRECTIVE_EFFECTS,
        "_reactions_raw": sentinels.CHILD_JOIN_DIRECTIVE_REACTIONS,
        "embargoes": sentinels.CHILD_JOIN_DIRECTIVE_EMBARGOES,
    }
    merged_fields = copy.deepcopy(parent_action_definition)
    for field_name, child_field_value in child_action_definition.items():
        merge_needed = (
            field_name in field_name_to_join_directive_sentinel
            and field_name_to_join_directive_sentinel[field_name] in child_action_definition["_join_directives"]
        )
        if merge_needed:
            _merge_action_field(
                merged_fields=merged_fields,
                child_fields=child_action_definition,
                parent_fields=parent_action_definition,
                field_name=field_name
            )
        else:
            merged_fields[field_name] = child_field_value
    merged_action_definition = internal_types.IntermediateActionDefinition(
        type=external_types.ConstructDiscriminator.ACTION,
        name=child_action_definition["name"],
        _template=child_action_definition["_template"],
        reserved=child_action_definition["reserved"],
        parent=None,  # Always `None` for an `IntermediateActionDefinition`, and eventually removed altogether
        roles=merged_fields["roles"],
        importance=merged_fields["importance"],
        tags=merged_fields["tags"],
        gloss=merged_fields["gloss"],
        report=merged_fields["report"],
        _conditions_raw=merged_fields["_conditions_raw"],
        scratch=merged_fields["scratch"],
        _effects_raw=merged_fields["_effects_raw"],
        _reactions_raw=merged_fields["_reactions_raw"],
        saliences=merged_fields["saliences"],
        associations=merged_fields["associations"],
        embargoes=merged_fields["embargoes"],
    )
    return merged_action_definition


def _merge_action_field(
    merged_fields: dict[str, Any],
    child_fields: dict[str, Any],
    parent_fields: dict[str, Any],
    field_name: str
) -> None:
    """Merges the given field using material from the respective parent and child actions.

    Args:
        merged_fields: The dictionary of merged fields that is being constructed.
        child_fields: The intermediate definition for the action that is inheriting material,
            treated as a dictionary of fields to simplify typing concerns.
        parent_fields: Intermediate definition for the action from which material is being inherited,
            treated as a dictionary of fields to simplify typing concerns.
        field_name: The name of the field to merge.

    Returns:
        Nothing. Mutates `merged_fields` in place.

    Raises:
        VivCompileError: Incompatible field values in the respective definitions.
    """
    both_dicts = isinstance(child_fields.get(field_name), dict) and isinstance(parent_fields.get(field_name), dict)
    both_lists = isinstance(child_fields.get(field_name), list) and isinstance(parent_fields.get(field_name), list)
    if field_name == "tags":  # Special case, since it's technically a dict (list expression)
        merged_fields[field_name]["value"] += child_fields[field_name]["value"]
    elif both_dicts:
        merged_fields[field_name].update(child_fields[field_name])
    elif both_lists:
        merged_fields[field_name] += child_fields[field_name]
    else:
        # Shouldn't be possible, but we'll keep it here as belt and suspenders
        raise errors.VivCompileError(
            f"Cannot join field '{field_name}' in action '{child_fields['name']}': "
            f"expected dict or list for both, got {type(child_fields[field_name]).__name__} "
            f"and {type(parent_fields[field_name]).__name__}"
        )


def _handle_role_renaming_declarations(
    merged_action_definition: internal_types.IntermediateActionDefinition,
    parent_action_name: external_types.ActionName,
    parent_role_names: set[external_types.RoleName]
) -> None:
    """Modify the given action definition in place to honor all role-renaming declarations.

    A role-renaming declaration (e.g. `new_name<<old_name`) allows an author to rename a role that
    is inherited from a parent action. This entails not just updating the role's name in the child
    action, but also updating any references to the role in material inherited from the parent.

    Args:
        merged_action_definition: A merged action definition, meaning one that was produced by merging
            parent content into the definition of a child that inherits from the parent.
        parent_action_name: The name of the parent action.
        parent_role_names: A set containing all the role names in the parent role.

    Returns:
        None (the input definitions are modified in place).
    """
    # Build a mapping from old names to new names
    old_name_to_new_name: dict[external_types.RoleName, external_types.RoleName] = {}
    for role_definition in merged_action_definition["roles"].values():
        if role_definition["renames"]:
            new_role_name = role_definition['name']
            old_role_name = role_definition["renames"]
            if old_role_name not in parent_role_names:
                raise errors.VivCompileError(
                    f"Action '{merged_action_definition['name']}' attempts to rename role '{old_role_name}', but "
                    f"no such role is defined in the parent action '{parent_action_name}'"
                )
            old_name_to_new_name[old_role_name] = new_role_name
    # Delete the definitions that only specify renamings
    for new_role_name in old_name_to_new_name.values():
        del merged_action_definition['roles'][new_role_name]
    # If the dictionary is empty, there's no role-renaming declarations to handle and we can return now
    if not old_name_to_new_name:
        return
    # Otherwise, let's proceed. First, update the corresponding role definitions
    for role_definition in merged_action_definition["roles"].values():
        if role_definition["name"] in old_name_to_new_name:
            role_definition["name"] = old_name_to_new_name[role_definition["name"]]
    merged_action_definition["roles"] = {role['name']: role for role in merged_action_definition['roles'].values()}
    # Update any embargo `roles` fields, which contain bare role names (not references)
    for embargo in merged_action_definition["embargoes"]:
        if not embargo["roles"]:
            continue
        updated_roles_field = []
        for role_name in embargo["roles"]:
            updated_role_name = old_name_to_new_name.get(role_name, role_name)
            updated_roles_field.append(updated_role_name)
        embargo["roles"] = updated_roles_field
    # Recursively walk the action definition to update all applicable references
    _rewrite_role_references(ast_chunk=merged_action_definition, old_name_to_new_name=old_name_to_new_name)


def _rewrite_role_references(
    ast_chunk: Any,
    old_name_to_new_name: dict[external_types.RoleName, external_types.RoleName]
) -> Any:
    """Recurse over the given AST chunk to honor any role-renaming declarations captured in the given mapping.

    This function only updates reference AST nodes, for both entity references and symbol references,
    which it reaches by recursively visiting all dictionary values and list elements.

    Args:
        ast_chunk: The AST chunk to search.
        old_name_to_new_name: A mapping from old role names to new role names, as specified in all the
            role-renaming declarations contained in a given action definition that uses inheritance.

    Returns:
        The updated AST chunk, which will also be mutated in place. (Note: its shape will never change.)
    """
    # Recurse over a list value
    if isinstance(ast_chunk, list):
        for i, element in enumerate(ast_chunk):
            ast_chunk[i] = _rewrite_role_references(ast_chunk=element, old_name_to_new_name=old_name_to_new_name)
        return ast_chunk
    # If it's a dictionary, we may just have a reference...
    if isinstance(ast_chunk, dict):
        node_type = ast_chunk.get("type")
        # Rename an entity-reference anchor, if applicable, and recurse over its value
        is_reference = (
                node_type == external_types.ExpressionDiscriminator.ENTITY_REFERENCE or
                node_type == external_types.ExpressionDiscriminator.SYMBOL_REFERENCE
        )
        if is_reference:
            value = ast_chunk.get("value")
            name_of_reference_anchor_role = value["anchor"]
            if name_of_reference_anchor_role in old_name_to_new_name:
                value["anchor"] = old_name_to_new_name[name_of_reference_anchor_role]
            for i, path_component in enumerate(value["path"]):
                value["path"][i] = _rewrite_role_references(
                    ast_chunk=path_component,
                    old_name_to_new_name=old_name_to_new_name
                )
            return ast_chunk
        # Recurse over any other dictionary value
        for key, value in ast_chunk.items():
            ast_chunk[key] = _rewrite_role_references(ast_chunk=value, old_name_to_new_name=old_name_to_new_name)
        return ast_chunk
    # For any other kind of value, there's no need to recurse
    return ast_chunk



"""Types capturing external schemas shared across the larger Viv project.

This module specifies in Python the types associated with a public, stable schema for Viv
compiled content bundles. The schemas capture the compiler's emitted JSON shapes, which are
mirrored in the corresponding runtime type definitions, assuming both are using the same
schema version number. As such, the types here constitute a reliable contract between the
compiler and any Viv runtime.

Note that a canonical JSON schema is also stored in `viv_compiler/schemas/compiler.schema.json`,
and that this is generated at the same time the runtime schemas are generated, which enforces
agreement between them. The final step in compilation is to structurally validate the compiled
content bundle against this schema. When a content bundle is registered with the Viv runtime,
the runtime structurally validates it using its corresponding JSON schema. Note that this JSON
schema is generated from the TypeScript types contained in the Viv JavaScript runtime.
"""

from .content_types import *
from .dsl_types import *



"""Utility functions used internally within the validation module."""

from typing import Any

from viv_compiler import config, errors, external_types, utils


def get_all_action_role_names(action_definition: external_types.ActionDefinition) -> set[external_types.RoleName]:
    """Return a set containing the names of all roles associated with the given action definition.

    Args:
        action_definition: The action definition whose role names will be returned.

    Returns:
        A set containing the names of all roles associated with the given action definition.
    """
    return config.SPECIAL_ROLE_NAMES | set(action_definition['roles'])


def get_all_assigned_scratch_variable_names(ast_chunk: Any) -> list[external_types.VariableName]:
    """Return a list of the names of all scratch variables that are assigned anywhere in the given AST chunk.

    Args:
        ast_chunk: The full or partial AST to search for scratch variables being assigned.

    Returns:
        A list containing the names of all scratch variables that are assigned anywhere in the given AST chunk.
    """
    all_assigned_scratch_variable_names = set()
    all_assignments: list[external_types.AssignmentValue] = utils.get_all_expressions_of_type(
        expression_type=external_types.ExpressionDiscriminator.ASSIGNMENT,
        ast_chunk=ast_chunk
    )
    for assignment in all_assignments:
        # Move on if the effective prefix is not `@this.scratch.`
        lhs = assignment['left']
        if lhs['type'] != external_types.ExpressionDiscriminator.ENTITY_REFERENCE:
            continue
        if lhs['value']['local']:
            continue
        if lhs['value']['anchor'] != config.SCRATCH_VARIABLE_REFERENCE_ANCHOR:
            continue
        if lhs['value']['path'][0] != config.SCRATCH_VARIABLE_REFERENCE_PATH_PREFIX[0]:
            continue
        # Now determine if the LHS of the assignment is a bare scratch variable, i.e., an effective `@this.scratch.var`
        lhs_value: external_types.ReferenceValue = lhs['value']
        head, *tail = lhs_value['path'][1:]
        if tail:
            continue
        if head['type'] != external_types.ReferencePathComponentDiscriminator.REFERENCE_PATH_COMPONENT_PROPERTY_NAME:
            # Move on if we somehow have a pointer or lookup after an explicit `@this.scratch` (not idiomatic)
            continue
        all_assigned_scratch_variable_names.add(head['name'])
    return sorted(all_assigned_scratch_variable_names)


def get_all_referenced_scratch_variable_names(ast_chunk: Any) -> list[external_types.VariableName]:
    """Return a list of the names of all scratch variables that are referenced anywhere in the given AST chunk.

    Args:
        ast_chunk: The full or partial AST to search for scratch variables being referenced.

    Returns:
        A list containing the names of all scratch variables that are referenced anywhere in the given AST chunk.
    """
    all_referenced_scratch_variable_names = set()
    all_entity_references: list[external_types.ReferenceValue] = utils.get_all_expressions_of_type(
        expression_type=external_types.ExpressionDiscriminator.ENTITY_REFERENCE,
        ast_chunk=ast_chunk
    )
    for entity_reference in all_entity_references:
        # Move on if the effective prefix is not `@this.scratch.`
        if entity_reference['local']:
            continue
        if entity_reference['anchor'] != config.SCRATCH_VARIABLE_REFERENCE_ANCHOR:
            continue
        if not entity_reference['path']:  # Bare `@this`
            continue
        if entity_reference['path'][0] != config.SCRATCH_VARIABLE_REFERENCE_PATH_PREFIX[0]:
            continue
        # Now determine if the LHS of the assignment is a bare scratch variable, i.e., an effective `@this.scratch.var`
        head, *_ = entity_reference['path'][1:]
        if head['type'] != external_types.ReferencePathComponentDiscriminator.REFERENCE_PATH_COMPONENT_PROPERTY_NAME:
            # Move on if we somehow have a pointer or lookup after an explicit `@this.scratch` (not idiomatic)
            continue
        all_referenced_scratch_variable_names.add(head['name'])
    return sorted(all_referenced_scratch_variable_names)


def get_all_negated_expressions(ast_chunk: Any) -> list[external_types.Expression]:
    """
    Return every negated expression in the given AST chunk.

    Args:
        ast_chunk: The full or partial AST to search for negated expressions.

    Returns:
        A list containing all the negated Viv expressions in the given AST chunk.
    """
    negated_expressions = []
    if isinstance(ast_chunk, list):
        for element in ast_chunk:
            negated_expressions.extend(get_all_negated_expressions(ast_chunk=element))
    elif isinstance(ast_chunk, dict):
        if ast_chunk.get("negated"):
            negated_expressions.append(ast_chunk)
        for value in ast_chunk.values():
            negated_expressions.extend(get_all_negated_expressions(ast_chunk=value))
    return negated_expressions


def detect_cycle_in_graph(adjacency_graph: dict[str, set[str]]) -> list[str] | None:
    """Returns a list describing a directed cycle in the given adjacency graph, if there is one, else `None`.

    Args:
        adjacency_graph: A mapping from node names to the set of node names to which they have outgoing edges.

    Returns:
        A list describing a directed cycle in the given adjacency graph, if there is one, else `None`.
    """
    white, gray, black = 0, 1, 2
    color = {name: white for name in adjacency_graph}
    stack = []

    def dfs(u: str) -> list[str] | None:
        """Carry out depth-first search starting from the given node, returning any detected cycle, else `None`.

        Args:
            u: The node from which to conduct a depth-first search (in the adjacency graph).

        Returns:
             A list describing an encountered directed cycle, if any, else `None`.
        """
        color[u] = gray
        stack.append(u)
        for v in adjacency_graph[u]:
            if color[v] == white:
                detected_cycle = dfs(v)
                if detected_cycle:
                    return detected_cycle
            elif color[v] == gray:
                # Found a cycle
                if v in stack:
                    i = stack.index(v)
                    return stack[i:] + [v]
                else:
                    return [v, u, v]
        stack.pop()
        color[u] = black
        return None

    for name in adjacency_graph:
        if color[name] == white:
            cycle = dfs(name)
            if cycle:
                return cycle
    return None


def is_non_numeric_expression(expression: external_types.Expression) -> bool:
    """Returns `True` if the given expression will certainly evaluate to a non-numeric value, else `False`.

    Note that we ultimately can't know what kinds of values certain expressions will produce,
    since they may depend on the host application (e.g., an adapter-function call).

    Args:
        expression: The expression in question.

    Returns:
        `True` if the given expression will certainly evaluate to a non-numeric value, else `False`.
    """
    match expression['type']:
        case external_types.ExpressionDiscriminator.ACTION_RELATION:
            return True
        case external_types.ExpressionDiscriminator.ACTION_SEARCH:
            return True
        case external_types.ExpressionDiscriminator.CUSTOM_FUNCTION_CALL:
            return False
        case external_types.ExpressionDiscriminator.ASSIGNMENT:
            return True
        case external_types.ExpressionDiscriminator.ARITHMETIC_EXPRESSION:
            return False
        case external_types.ExpressionDiscriminator.BOOL:
            return True
        case external_types.ExpressionDiscriminator.CHANCE_EXPRESSION:
            return True
        case external_types.ExpressionDiscriminator.COMPARISON:
            return True
        case external_types.ExpressionDiscriminator.CONDITIONAL:
            return False
        case external_types.ExpressionDiscriminator.CONJUNCTION:
            return True
        case external_types.ExpressionDiscriminator.DISJUNCTION:
            return True
        case external_types.ExpressionDiscriminator.ENTITY_REFERENCE:
            return False
        case external_types.ExpressionDiscriminator.ENUM:
            return False
        case external_types.ExpressionDiscriminator.FLOAT:
            return False
        case external_types.ExpressionDiscriminator.INSPECTION:
            return True
        case external_types.ExpressionDiscriminator.INSCRIPTION:
            return True
        case external_types.ExpressionDiscriminator.INT:
            return False
        case external_types.ExpressionDiscriminator.LIST:
            return True
        case external_types.ExpressionDiscriminator.LOOP:
            return False
        case external_types.ExpressionDiscriminator.MEMBERSHIP_TEST:
            return True
        case external_types.ExpressionDiscriminator.MEMORY_CHECK:
            return True
        case external_types.ExpressionDiscriminator.NULL_TYPE:
            return True
        case external_types.ExpressionDiscriminator.OBJECT:
            return True
        case external_types.ExpressionDiscriminator.REACTION:
            return True
        case external_types.ExpressionDiscriminator.SIFTING:
            return True
        case external_types.ExpressionDiscriminator.STRING:
            return True
        case external_types.ExpressionDiscriminator.SYMBOL_REFERENCE:
            return False
        case external_types.ExpressionDiscriminator.TEMPLATE_STRING:
            return True
        case external_types.ExpressionDiscriminator.TROPE_FIT:
            return True
        case _:
            raise errors.VivCompileError(
                msg=f"Unexpected expression discriminator: '{expression['type']}'",
                source=expression.get('source'),
            )



"""Handles preliminary validation of certain aspects of intermediate construct definitions,
to support phases of postprocessing that rely on those aspects being well-formed.
"""

__all__ = [
    "prevalidate_role_names",
    "prevalidate_construct_names",
    "prevalidate_action_inheritance",
    "prevalidate_initiator_roles",
    "prevalidate_roles",
    "prevalidate_action_selectors"
]

from viv_compiler import config, errors, external_types, internal_types, sentinels, utils
from ._utils import detect_cycle_in_graph


def prevalidate_role_names(
    construct_type: external_types.ConstructDiscriminator,
    construct_name: str,
    role_definitions: list[external_types.RoleDefinition],
    source: external_types.SourceAnnotations | None = None
) -> None:
    """Ensure that there are no duplicate names among the given role definitions for the given construct.

    Args:
        construct_type: The type of construct associated with these roles.
        construct_name: The name of the construct associated with these roles.
        role_definitions: The role definitions to check for duplicates.
        source: If applicable, source annotations for the construct.

    Returns:
        None.

    Raises:
        VivCompileError: A duplicate name was detected among the given role definitions.
    """
    for role_definition in role_definitions:
        n_roles_with_that_name = 0
        for other_role_definition in role_definitions:
            if other_role_definition['name'] == role_definition["name"]:
                n_roles_with_that_name += 1
                if n_roles_with_that_name > 1:
                    raise errors.VivCompileError(
                        f"{utils.CONSTRUCT_LABEL[construct_type]} '{construct_name}' "
                        f"has duplicate role: '{role_definition['name']}'",
                        source=source
                    )


def prevalidate_construct_names(combined_ast: internal_types.CombinedAST) -> None:
    """Ensure that there are no duplicate names among the units in the given combined AST.

    More specifically, there must not be a duplicate name *within* a unit type. For example,
    there may not be two actions with the same name, but it's fine for an action and a trope
    to share the same name.

    Args:
        combined_ast: An abstract syntax tree produced by the Visitor class, integrating the
            respective ASTs of any included files.

    Returns:
        None.

    Raises:
        VivCompileError: A duplicate name was detected among the units in the given combined AST.
    """
    # Detect duplicate action names
    action_names = [action_definition['name'] for action_definition in combined_ast['actions']]
    for action_name in action_names:
        if action_names.count(action_name) > 1:
            raise errors.VivCompileError(f"Duplicate action name: '{action_name}'")
    # Detect duplicate action-selector names
    action_selector_names = [selector_definition['name'] for selector_definition in combined_ast['actionSelectors']]
    for action_selector_name in action_selector_names:
        if action_selector_names.count(action_selector_name) > 1:
            raise errors.VivCompileError(f"Duplicate action-selector name: '{action_selector_name}'")
    # Detect duplicate plan names
    plan_names = [plan_definition['name'] for plan_definition in combined_ast['plans']]
    for plan_name in plan_names:
        if plan_names.count(plan_name) > 1:
            raise errors.VivCompileError(f"Duplicate plan name: '{plan_name}'")
    # Detect duplicate plan-selector names
    plan_selector_names = [selector_definition['name'] for selector_definition in combined_ast['planSelectors']]
    for plan_selector_name in plan_selector_names:
        if plan_selector_names.count(plan_selector_name) > 1:
            raise errors.VivCompileError(f"Duplicate plan-selector name: '{plan_selector_name}'")
    # Detect duplicate query names
    query_names = [query_definition['name'] for query_definition in combined_ast['queries']]
    for query_name in query_names:
        if query_names.count(query_name) > 1:
            raise errors.VivCompileError(f"Duplicate query name: '{query_name}'")
    # Detect duplicate sifting-pattern names
    sifting_pattern_names = [pattern_definition['name'] for pattern_definition in combined_ast['siftingPatterns']]
    for sifting_pattern_name in sifting_pattern_names:
        if sifting_pattern_names.count(sifting_pattern_name) > 1:
            raise errors.VivCompileError(f"Duplicate sifting-pattern name: '{sifting_pattern_name}'")
    # Detect duplicate trope names
    trope_names = [trope_definition['name'] for trope_definition in combined_ast['tropes']]
    for trope_name in trope_names:
        if trope_names.count(trope_name) > 1:
            raise errors.VivCompileError(f"Duplicate trope name: '{trope_name}'")


def prevalidate_action_inheritance(combined_ast: internal_types.CombinedAST) -> None:
    """Ensure that action definitions only use inheritance fields when a parent action is specified.

    If this validation check passes, the given combined AST is ready for inheritance to be handled.

    Args:
        combined_ast: An abstract syntax tree produced by the Visitor class, integrating the
            respective ASTs of any included files.

    Returns:
        None.

    Raises:
        VivCompileError: The given combined AST did not pass validation.
    """
    for action_definition in combined_ast["actions"]:
        # Validate use of `join` operator
        if "_join_directives" in action_definition and not action_definition["parent"]:
            raise errors.VivCompileError(
                f"Action '{action_definition['name']}' uses 'join' operator but declares no parent"
            )
        # Validate use of role renaming
        role_definitions = action_definition.get('roles', {}).values()
        for role_definition in role_definitions:
            if role_definition["renames"]:
                if not action_definition["parent"]:
                    raise errors.VivCompileError(
                        f"Action '{action_definition['name']}' renames a role but declares no parent"
                    )
                if sentinels.CHILD_JOIN_DIRECTIVE_ROLES not in action_definition.get("_join_directives", []):
                    raise errors.VivCompileError(
                        f"Action '{action_definition['name']}' uses role renaming but does not 'join roles'"
                    )


def prevalidate_initiator_roles(combined_ast: internal_types.CombinedAST) -> None:
    """Ensure that each of the action and selector definitions in the given AST
    has a single initiator role with no casting pool.

    Args:
        combined_ast: An abstract syntax tree produced by the Visitor class, integrating the
            respective ASTs of any included files.

    Returns:
        None.

    Raises:
        VivCompileError: An action definition has multiple initiators, or has an initiator with a casting pool.
    """
    for construct_definition in utils.get_all_construct_definitions(combined_ast=combined_ast):
        if construct_definition['type'] not in config.CONSTRUCT_TYPES_USING_INITIATOR_ROLES:
            continue
        # Make sure there is exactly one initiator role
        all_initiator_roles = [
            role for role in construct_definition['roles'].values() if utils.is_initiator_role(role_definition=role)
        ]
        if len(all_initiator_roles) == 0:
            raise errors.VivCompileError(
                f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                f"'{construct_definition['name']}' has no initiator role (must have exactly one)"
            )
        elif len(all_initiator_roles) > 1:
            raise errors.VivCompileError(
                f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                f"'{construct_definition['name']}' has multiple initiator roles (must have exactly one)"
            )
        # Make sure the initiator role does not have a casting-pool directive. We prohibit
        # this because the initiator role is always cast first.
        initiator_role_definition = all_initiator_roles[0]
        if initiator_role_definition['pool']:
            raise errors.VivCompileError(
                f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                f"'{construct_definition['name']}' has initiator role '{initiator_role_definition['name']}' "
                f"with a casting-pool directive, which is not allowed for initiator roles"
            )


def prevalidate_roles(combined_ast: internal_types.CombinedAST) -> None:
    """Detect any issues that would preclude the construction of role-dependency forests
    for the constructs in the given combined AST.

    Args:
        combined_ast: A combined AST with wrapped expressions and actions that have the `initiator` field set.

    Returns:
        None.

    Raises:
        VivCompileError: At least one construct definition did not pass validation.
    """
    for construct_definition in utils.get_all_construct_definitions(combined_ast=combined_ast):
        # Detect any reference to an undefined role. This validation *must* occur prior
        # to condition attribution and construction of the role-dependency forest.
        _detect_reference_to_undefined_role(construct_definition=construct_definition)
        # Detect cases of a casting pool referencing an optional role
        _detect_casting_pool_optional_role_anchor(construct_definition=construct_definition)
        # Detect cycles across the casting pools for the construct roles
        _detect_casting_pool_cycle(construct_definition=construct_definition)
        # Detect cases of a condition referencing multiple optional roles
        _detect_condition_referencing_multiple_optional_roles(construct_definition=construct_definition)


def _detect_reference_to_undefined_role(construct_definition: internal_types.IntermediateConstructDefinition) -> None:
    """Ensure that the given construct definition has no references to any undefined role.

    Args:
        construct_definition: An intermediate construct definition.

    Returns:
        None.

    Raises:
        VivCompileError: A reference to an undefined role was detected.
    """
    field_label = {field_name: field_name for field_name in construct_definition}
    field_label["_conditions_raw"] = "conditions"
    field_label["_effects_raw"] = "effects"
    field_label["_reactions_raw"] = "effects"
    for field_name, field_value in construct_definition.items():
        for role_name in utils.get_all_referenced_roles(ast_chunk=field_value):
            if role_name not in set(construct_definition['roles']) | config.SPECIAL_ROLE_NAMES:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
                    f"references undefined role '{role_name}' in '{field_label[field_name]}' field"
                )


def _detect_casting_pool_optional_role_anchor(
    construct_definition: internal_types.IntermediateConstructDefinition
) -> None:
    """Throws an error if any casting-pool directive in the given construct's roles is anchored in an optional role.

    Args:
        construct_definition: An intermediate construct definition. If it's an action, the `initiator`
            field must be present.

    Returns:
        None. Only returns if no pool is anchored in an optional role.

    Raises:
        VivCompileError: If a pool is anchored in an optional role.
    """
    optional_role_names = {role['name'] for role in construct_definition['roles'].values() if role['min'] == 0}
    for role_definition in construct_definition['roles'].values():
        if role_definition['pool']:
            for anchor_role_name in utils.get_all_referenced_roles(ast_chunk=role_definition["pool"]):
                if anchor_role_name in optional_role_names:
                    raise errors.VivCompileError(
                        f"Casting-pool directive for role '{role_definition['name']}' in "
                        f"{utils.CONSTRUCT_LABEL[construct_definition['type']].lower()} "
                        f"'{construct_definition['name']}' references an optional role "
                        f"in its casting-pool directive: '{anchor_role_name}'"
                    )


def _detect_casting_pool_cycle(construct_definition: internal_types.IntermediateConstructDefinition) -> None:
    """Throws an error if the casting-pool directives for the given action's roles form a cycle.

    Args:
        construct_definition: An intermediate action definition whose 'initiator' field has already been set.

    Returns:
        None. Only returns if there's no cycle.

    Raises:
        VivCompileError: If a cycle is detected.
    """
    # Build adjacency graph
    required_non_initiator_roles = [
        role for role in construct_definition['roles'].values()
        if role['min'] > 0 and not utils.is_initiator_role(role_definition=role)
    ]
    required_non_initiator_role_names = {role['name'] for role in required_non_initiator_roles}
    adjacency_graph = {role['name']: set() for role in required_non_initiator_roles}
    for role in required_non_initiator_roles:
        if role['pool']:
            for anchor_role_name in utils.get_all_referenced_roles(ast_chunk=role["pool"]):
                if anchor_role_name == role['name']:
                    raise errors.VivCompileError(
                        f"Role '{role['name']}' in "
                        f"{utils.CONSTRUCT_LABEL[construct_definition['type']].lower()} "
                        f"'{construct_definition['name']}' references itself in its casting-pool directive "
                        f"(prohibited because it cannot be cast without first being cast)"
                    )
                if anchor_role_name in required_non_initiator_role_names:
                    adjacency_graph[anchor_role_name].add(role['name'])
    # Attempt to detect a cycle
    cycle = detect_cycle_in_graph(adjacency_graph=adjacency_graph)
    if cycle:
        cycle_str = "'" + "' > '".join(cycle) + "'"
        raise errors.VivCompileError(
            f"Cycle detected among casting-pool anchors in "
            f"{utils.CONSTRUCT_LABEL[construct_definition['type']].lower()} "
            f"'{construct_definition['name']}': {cycle_str} "
            f"(prohibited because such a cycle can never be resolved)"
        )


def _detect_condition_referencing_multiple_optional_roles(
    construct_definition: internal_types.IntermediateConstructDefinition
) -> None:
    """Ensure that none of the given construct definition's conditions references multiple optional roles.

    We prohibit this because optional roles are cast one at a time, with no backtracking. As such, it's
    currently not possible to evaluate a condition referencing two optional roles, since that would require
    considering the combinatorial space of their respective candidates, which entails backtracking.

    Args:
        construct_definition: An intermediate construct definition.

    Returns:
        None.

    Raises:
        VivCompileError: One or more of the given construct definition's conditions references multiple optional roles.
    """
    for condition in construct_definition['_conditions_raw']:
        optional_role_references = []
        for reference in condition['references']:
            if construct_definition['roles'][reference]['min'] < 1:
                optional_role_references.append(reference)
            if len(optional_role_references) > 1:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                    f"'{construct_definition['name']}' has condition that references multiple optional "
                    f"roles ({', '.join(optional_role_references)}), but currently a condition may only "
                    f"reference at most one optional role"
                )


def prevalidate_action_selectors(combined_ast: internal_types.CombinedAST) -> None:
    """Prevalidate all action selectors.

    Args:
        combined_ast: A combined AST as produced by inheritance processing.

    Returns:
        None.
    """
    for action_selector_definition in combined_ast["actionSelectors"]:
        _detect_missing_action_selector_initiator(action_selector_definition=action_selector_definition)
        _detect_undefined_action_selector_candidate(
            action_selector_definition=action_selector_definition,
            combined_ast=combined_ast
        )
    _detect_action_selector_candidate_cycle(combined_ast=combined_ast)


def _detect_missing_action_selector_initiator(
        action_selector_definition: internal_types.IntermediateActionSelectorDefinition
) -> None:
    """Ensure that the given action selector has a valid (potentially implicit) initiator role.

    Args:
        action_selector_definition: An intermediate action-selector definition.

    Returns:
        None.

    Raises:
        VivCompileError: The action selector defines a `roles` field, but does not specify an initiator role.
    """
    if action_selector_definition["roles"]:
        role_definitions = action_selector_definition["roles"].values()
        if not any(utils.is_initiator_role(role_definition=role) for role in role_definitions):
            raise errors.VivCompileError(
                f"Action selector '{action_selector_definition['name']}' has a 'roles' field, "
                f"but fails to specify an initiator"
            )


def _detect_undefined_action_selector_candidate(
    action_selector_definition: internal_types.IntermediateActionSelectorDefinition,
    combined_ast: internal_types.CombinedAST
) -> None:
    """Ensure that all the given action selector's candidates are defined actions and/or action selectors.

    Args:
        action_selector_definition: An intermediate action-selector definition.
        combined_ast: A combined AST as produced by inheritance processing.

    Returns:
        None.

    Raises:
        VivCompileError: The action selector has an undefined candidate.
    """
    for candidate in action_selector_definition["candidates"]:
        if candidate["isSelector"]:
            if not any(selector['name'] == candidate['name'] for selector in combined_ast["actionSelectors"]):
                raise errors.VivCompileError(
                    f"Action selector '{action_selector_definition['name']}' targets undefined "
                    f"action selector: '{candidate['name']}'"
                )
        elif not any(action['name'] == candidate['name'] for action in combined_ast["actions"]):
            raise errors.VivCompileError(
                f"Action selector '{action_selector_definition['name']}' targets "
                f"undefined action: '{candidate['name']}'"
            )


def _detect_action_selector_candidate_cycle(combined_ast: internal_types.CombinedAST) -> None:
    """Throws an error if there is a cycle among the action-selector definitions in the given combined AST.

    A selector can yield a cycle when it targets itself, or when it targets another selector
    that directly or indirectly ends up targeting the original selector.

    Args:
        combined_ast: A combined AST as produced by inheritance processing.

    Returns:
        None. Only returns if there's no cycle.
    """
    # Build adjacency graph
    adjacency_graph = {selector_definition['name']: set() for selector_definition in combined_ast["actionSelectors"]}
    for selector_definition in combined_ast["actionSelectors"]:
        for candidate in selector_definition["candidates"]:
            if candidate["isSelector"]:
                if candidate["name"] == selector_definition["name"]:
                    raise errors.VivCompileError(
                        f"Action selector '{selector_definition['name']}' directly targets itself "
                        f"(prohibited because it would yield an infinite loop during targeting)"
                    )
                adjacency_graph[selector_definition['name']].add(candidate["name"])
    # Attempt to select an indirect cycle
    cycle = detect_cycle_in_graph(adjacency_graph=adjacency_graph)
    if cycle:
        cycle_str = "'" + "' > '".join(cycle) + "'"
        raise errors.VivCompileError(
            f"Cycle detected among action selectors: {cycle_str} "
            f"(prohibited because such cycles make selector targeting impossible to resolve)"
        )



"""Types associated with the lower-level concerns represented in the Viv DSL's abstract syntax trees."""

from __future__ import annotations

from enum import StrEnum
from typing import TYPE_CHECKING, Literal, TypeAlias, TypedDict, Union

if TYPE_CHECKING:
    # Imported for static checking only; referenced by string at runtime to avoid cycles
    from .content_types import (
        ActionName,
        QueryName,
        PlanName,
        ReactionTargetConstructDiscriminator,
        RoleName,
        SelectorName,
        SiftingPatternName,
        TropeName
    )


# A Viv expression
Expression: TypeAlias = Union[
    "ActionRelation",
    "ActionSearch",
    "CustomFunctionCall",
    "ArithmeticExpression",
    "Assignment",
    "BoolField",
    "ChanceExpression",
    "Comparison",
    "Conditional",
    "Conjunction",
    "Disjunction",
    "EntityReference",
    "Enum",
    "FloatField",
    "Inscription",
    "Inspection",
    "IntField",
    "ListField",
    "Loop",
    "MembershipTest",
    "MemoryCheck",
    "NullField",
    "ObjectField",
    "Reaction",
    "Sifting",
    "StringField",
    "SymbolReference",
    "TemplateStringField",
    "TropeFit",
]


# Enum containing discriminators for each Viv expression type
class ExpressionDiscriminator(StrEnum):
    ACTION_RELATION = "actionRelation"
    ACTION_SEARCH = "actionSearch"
    ASSIGNMENT = "assignment"
    ARITHMETIC_EXPRESSION = "arithmeticExpression"
    BOOL = "bool"
    CHANCE_EXPRESSION = "chanceExpression"
    COMPARISON = "comparison"
    CONDITIONAL = "conditional"
    CONJUNCTION = "conjunction"
    CUSTOM_FUNCTION_CALL = "customFunctionCall"
    DISJUNCTION = "disjunction"
    ENTITY_REFERENCE = "entityReference"
    ENUM = "enum"
    FLOAT = "float"
    INSPECTION = "inspection"
    INSCRIPTION = "inscription"
    INT = "int"
    LIST = "list"
    LOOP = "loop"
    MEMBERSHIP_TEST = "membershipTest"
    MEMORY_CHECK = "memoryCheck"
    NULL_TYPE = "nullType"
    OBJECT = "object"
    REACTION = "reaction"
    SIFTING = "sifting"
    STRING = "string"
    SYMBOL_REFERENCE = "symbolReference"
    TEMPLATE_STRING = "templateString"
    TROPE_FIT = "tropeFit"


class SourceAnnotatedExpression(TypedDict):
    """Mixin for all expression types that specifies metadata about the expression source code."""
    # Annotations on the expression specifying its original source code and position
    # in a source file. This will be `None` in certain expressions, such as default
    # values for optional fields, that do not originate in source code.
    source: SourceAnnotations | None


class SourceAnnotations(TypedDict):
    """An object containing annotations on a Viv expression, specifying its original
    source code and position in a source file.
    """
    # Absolute path to the source file including the expression
    filePath: str
    # Line number (1-based) at the start of the expression
    line: int
    # Column number (1-based) at the start of the expression
    column: int
    # Line number (1-based) at the end of the expression
    endLine: int
    # Column number (1-based) at the end of the expression
    endColumn: int
    # The source code for the expression
    code: str


# Mixin for expression types that may be negated
class NegatableExpression(TypedDict):
    """Mixin for expression types that may be negated."""
    # Whether to negate the result of the expression. Only present when `True`.
    negated: bool


# Mixin for expression components that may fail safely
class FailSafeComponent(TypedDict):
    """Mixin for expression components that may fail safely."""
    # Whether the expression component should fail safely by converting a nullish evaluation
    # into the falsy 'eval fail-safe signal'. This allows for references whose (intermediate)
    # references may be undefined, as in `@foo.bar?.baz`. Only present when `True`.
    failSafe: bool


class ActionRelation(SourceAnnotatedExpression, NegatableExpression):
    """A Viv action relation.

    This is a kind of expression which evaluates to `true` if a specified relation holds between two actions.
    """
    # Discriminator for a Viv action relation
    type: Literal[ExpressionDiscriminator.ACTION_RELATION]
    # The actual expression value
    value: ActionRelationValue


class ActionRelationValue(TypedDict):
    """The actual expression value for a Viv action relation."""
    # An expression whose evaluation will be used as the left operand in relation test
    left: Expression
    # The relation operator
    operator: ActionRelationOperator
    # An expression whose evaluation will be used as the right operand in relation test
    right: Expression


# Enum containing the action-relation operators supported by Viv
ActionRelationOperator = Literal["caused", "preceded", "triggered"]


class CustomFunctionCall(SourceAnnotatedExpression, NegatableExpression):
    """A Viv custom-function call, which parameterizes a call to some function that must
    be registered in the host application's Viv adapter.

    For instance, a Viv author might specify a function call in an action such as
    `~transport(@person.id, @destination.id)`, in which case there must be a function
    `transport()` exposed in the adapter. The Viv runtime confirms the existence of all
    referenced function names during adapter initialization.
    """
    # Discriminator for a Viv adapter function call
    type: Literal[ExpressionDiscriminator.CUSTOM_FUNCTION_CALL]
    # The actual expression value
    value: CustomFunctionCallValue


class CustomFunctionCallValue(FailSafeComponent):
    """The actual expression value for a Viv custom-function call."""
    # The name of the custom function. There must be a function stored in the host
    # application's Viv adapter, via a key by this same name.
    name: CustomFunctionName
    # An ordered list of Viv expressions whose evaluations will be passed as
    # arguments to the function, in that same order.
    args: list[Expression]


# The name for a custom function targeted by a Viv custom-function call
CustomFunctionName: TypeAlias = str


class ArithmeticExpression(SourceAnnotatedExpression):
    """A Viv arithmetic expression, which accepts two numeric operands and evaluates to a number."""
    # Discriminator for a Viv arithmetic expression
    type: Literal[ExpressionDiscriminator.ARITHMETIC_EXPRESSION]
    # The actual expression value
    value: ArithmeticExpressionValue


class ArithmeticExpressionValue(TypedDict):
    """The actual expression value for a Viv arithmetic expression."""
    # An expression whose evaluation will be used as the left operand in the arithmetic expression
    left: Expression
    # The arithmetic operator
    operator: ArithmeticOperator
    # An expression whose evaluation will be used as the right operand in the arithmetic expression
    right: Expression


# Enum containing the arithmetic operators supported by Viv
ArithmeticOperator = Literal["+", "-", "*", "/"]


class Assignment(SourceAnnotatedExpression):
    """A Viv assignment (or update)."""
    # Discriminator for a Viv assignment
    type: Literal[ExpressionDiscriminator.ASSIGNMENT]
    # The actual expression value
    value: AssignmentValue


class AssignmentValue(TypedDict):
    """The actual expression value for a Viv assignment."""
    # An expression whose evaluation will be used as the left operand in the assignment/update
    left: EntityReference | SymbolReference
    # The assignment/update operator
    operator: AssignmentOperator
    # An expression whose evaluation will be used as the right operand in the assignment/update. Note
    # that for assignments that update persistent entity data, the value will always be proactively
    # dehydrated, such that all entity data included in the value will be converted into the associated
    # entity ID. We do this to prevent several potential issues, and the data can be rehydrated later on.
    right: Expression


# Enum containing the Viv assignment (and update) operators
AssignmentOperator = Literal["=", "+=", "-=", "*=", "/=", "append", "remove"]


class BoolField(SourceAnnotatedExpression):
    """A Viv boolean."""
    # Discriminator for a Viv boolean
    type: Literal[ExpressionDiscriminator.BOOL]
    # The boolean literal to which this expression will evaluate
    value: bool


class ChanceExpression(SourceAnnotatedExpression):
    """A Viv chance expression.

    This is a kind of condition that evaluates to True if the specified probability value (a number
    between 0.0 and 1.0) exceeds a pseudorandom number generated by the Viv interpreter.
    """
    # Discriminator for a Viv chance expression
    type: Literal[ExpressionDiscriminator.CHANCE_EXPRESSION]
    # The specified probability, which the compiler guarantees to be a number in the range [0, 1]
    value: float


class Comparison(SourceAnnotatedExpression, NegatableExpression):
    """A Viv comparison, whereby two values are compared using a comparator."""
    # Discriminator for a Viv comparison
    type: Literal[ExpressionDiscriminator.COMPARISON]
    # The actual expression value
    value: ComparisonValue


class ComparisonValue(TypedDict):
    """The actual expression value for a Viv comparison."""
    # An expression whose evaluation will serve as the left operand in the comparison
    left: Expression
    # The comparison operator
    operator: Comparator
    # An expression whose evaluation will serve as the right operand in the comparison
    right: Expression


# Enum containing the Viv comparison operators
Comparator = Literal["==", ">", ">=", "<", "<=", "!="]


class Conditional(SourceAnnotatedExpression):
    """A Viv conditional expression, allowing for branching based on the value of a test."""
    # Discriminator for a Viv conditional
    type: Literal[ExpressionDiscriminator.CONDITIONAL]
    # The actual expression value
    value: ConditionalValue


class ConditionalValue(TypedDict):
    """The actual expression value for a Viv conditional."""
    # Branches representing the `if` and `elif` clauses in this conditional expression
    branches: list[ConditionalBranch]
    # If an author has provided an alternative body (via an `else` clause), a list
    # of expressions that will be evaluated/executed should the condition not hold.
    alternative: list[Expression] | None


class ConditionalBranch(TypedDict):
    """A Viv conditional branch, representing an `if` or `elif` clause."""
    # The condition that will be tested, which holds if its evaluation is truthy
    condition: Expression
    # A list of expressions that will be evaluated/executed should the condition hold
    consequent: list[Expression]


class Conjunction(SourceAnnotatedExpression, NegatableExpression):
    """A Viv conjunction.

    This kind of expression takes multiple expressions as operands and evaluates to
    `True` if and only if all the respective expressions evaluate to truthy values.
    """
    # Discriminator for a Viv conjunction
    type: Literal[ExpressionDiscriminator.CONJUNCTION]
    # The actual expression value
    value: ConjunctionValue


class ConjunctionValue(TypedDict):
    """The actual expression value for a Viv conjunction."""
    # A list of expressions that will be evaluated in turn to determine the result
    # of the conjunction. Note that the interpreter stops evaluating as soon as a
    # falsy evaluation is encountered.
    operands: list[Expression]


class Disjunction(SourceAnnotatedExpression, NegatableExpression):
    """A Viv disjunction.

    This kind which takes multiple expressions and evaluates to `True` if and only
    if at least one of the respective expressions evaluate to a truthy value.
    """
    # Discriminator for a Viv disjunction
    type: Literal[ExpressionDiscriminator.DISJUNCTION]
    # The actual expression value
    value: DisjunctionValue


class DisjunctionValue(TypedDict):
    """The actual expression value for a Viv disjunction."""
    # A list of expressions that will be evaluated in turn to determine the result
    # of the disjunction. Note that the interpreter stops evaluating as soon as a
    # truthy evaluation is encountered.
    operands: list[Expression]


class EntityReference(SourceAnnotatedExpression, NegatableExpression):
    """A Viv entity reference, structured as an anchor name and a (possibly empty) path to a specific property value.

    Usually, the property is on the anchor entity, but this is not the case if the reference contains
    a pointer. For instance, `@person.boss->boss->traits.cruel` would return the value stored at the
    path `traits.cruel` on the boss of the boss of the entity cast in the anchor role `@person`.

    Note that the compiler prevents an author from anchoring an entity reference in a symbol.

    Also note that references anchored in scratch variables, e.g. `$@foo.bar.baz`, are compiled
    to entity references -- this is because `$` is really just syntactic sugar for `@this.scratch.`,
    with the next sigil indicating the type of the scratch variable.
    """
    # Discriminator for a Viv entity reference
    type: Literal[ExpressionDiscriminator.ENTITY_REFERENCE]
    # The actual expression value
    value: ReferenceValue


class SymbolReference(SourceAnnotatedExpression, NegatableExpression):
    """A Viv symbol reference, structured as an anchor name and a (possibly empty) path to a specific property value."""
    # Discriminator for a Viv symbol reference
    type: Literal[ExpressionDiscriminator.SYMBOL_REFERENCE]
    # The actual expression value
    value: ReferenceValue


class ReferenceValue(FailSafeComponent):
    """The actual expression value for a Viv entity reference or symbol reference."""
    # The name anchoring this reference
    anchor: RoleName | VariableName
    # If applicable, the path to a specified property value. If the reference is just to the entity
    # or symbol itself, this will be an empty list. Otherwise, it will specify a path to a specific
    # property value, either on that entity or symbol or another entity (via a pointer).
    path: list[ReferencePathComponent]
    # Whether the anchor is a local variable. This is a common pattern when an author loops over a group role.
    local: bool
    # Whether the anchor is marked as a group role, in which case it will evaluate
    # to an array containing the bindings for the role.
    group: bool


class ReferencePathComponentPropertyName(FailSafeComponent):
    """A component of a Viv reference path specifying a property to access."""
    # Discriminator for a property-name reference path component
    type: Literal[ReferencePathComponentDiscriminator.REFERENCE_PATH_COMPONENT_PROPERTY_NAME]
    # The name of the property to access
    name: str


class ReferencePathComponentPointer(FailSafeComponent):
    """A component of a Viv reference path specifying a pointer to dereference."""
    # Discriminator for a pointer reference path component
    type: Literal[ReferencePathComponentDiscriminator.REFERENCE_PATH_COMPONENT_POINTER]
    # The name of the property to access in the entity data of the entity targeted by the pointer
    propertyName: str


class ReferencePathComponentLookup(FailSafeComponent):
    """A component of a Viv reference path specifying a property lookup or an array access.

    The key/index can be specified by an arbitrary Viv expression.
    """
    # Discriminator for a lookup reference path component
    type: Literal[ReferencePathComponentDiscriminator.REFERENCE_PATH_COMPONENT_LOOKUP]
    # An expression that should evaluate to a valid property key (string or integer)
    key: Expression


class ReferencePathComponentDiscriminator(StrEnum):
    """Enum containing discriminators for the possible reference path components."""
    # Discriminator for a property-name reference path component
    REFERENCE_PATH_COMPONENT_PROPERTY_NAME = "referencePathComponentPropertyName"
    # Discriminator for a pointer reference path component
    REFERENCE_PATH_COMPONENT_POINTER = "referencePathComponentPointer"
    # Discriminator for a lookup reference path component
    REFERENCE_PATH_COMPONENT_LOOKUP = "referencePathComponentLookup"


# A component in a Viv reference path
ReferencePathComponent = (
    ReferencePathComponentPropertyName
    | ReferencePathComponentPointer
    | ReferencePathComponentLookup
)


class Enum(SourceAnnotatedExpression):
    """A Viv enum.

    Enums are resolved at runtime using the host application's Viv adapter.
    """
    # Discriminator for a Viv enum
    type: Literal[ExpressionDiscriminator.ENUM]
    # The actual expression value
    value: EnumValue


class EnumValue(TypedDict):
    """The actual expression value for a Viv enum."""
    # The name of the enum. This must be resolvable by the host application at runtime.
    name: EnumName
    # Whether to flip the sign of a numeric value associated with the enum.
    minus: bool


# A unique label for an enum value
EnumName: TypeAlias = str


class FloatField(SourceAnnotatedExpression):
    """A Viv floating-point number."""
    # Discriminator for a Viv floating-point number
    type: Literal[ExpressionDiscriminator.FLOAT]
    # The float literal to which this expression will evaluate
    value: float


class Inspection(SourceAnnotatedExpression):
    """A Viv inspection.

    This kind of expression causes a character to *inspect* an item, which
    in turn may lead them to learn about actions inscribed in the item.
    """
    # Discriminator for a Viv inspection
    type: Literal[ExpressionDiscriminator.INSPECTION]
    # The actual expression value
    value: InspectionValue


class InspectionValue(TypedDict):
    """The actual expression value for a Viv inspection."""
    # An expression whose evaluation will be treated as the character operand in an inspection
    character: Expression
    # An expression whose evaluation will be treated as the item operand in an inspection
    item: Expression


class Inscription(SourceAnnotatedExpression):
    """A Viv inscription.

    This kind of expression causes an item to be inscribed with knowledge about a given
    action. If a character later inspects the item, they will learn about the action.
    """
    # Discriminator for a Viv inscription
    type: Literal[ExpressionDiscriminator.INSCRIPTION]
    # The actual expression value
    value: InscriptionValue


class InscriptionValue(TypedDict):
    """The actual expression value for a Viv inscription."""
    # An expression whose evaluation will be treated as the item operand in an inscription
    item: Expression
    # An expression whose evaluation will be treated as the action operand in an inscription
    action: Expression


class IntField(SourceAnnotatedExpression):
    """A Viv integer."""
    # Discriminator for a Viv integer
    type: Literal[ExpressionDiscriminator.INT]
    # The integer literal to which this expression will evaluate
    value: int


class ListField(SourceAnnotatedExpression):
    """A Viv list, defined as an ordered list of Viv expressions.

    Once evaluated, the result will contain the respective evaluations of the expressions, in that same order.
    """
    # Discriminator for a Viv list
    type: Literal[ExpressionDiscriminator.LIST]
    # The actual expression value
    value: list[Expression]


class Loop(SourceAnnotatedExpression):
    """A Viv loop, allowing for iteration over some iterable value."""
    # Discriminator for a Viv loop
    type: Literal[ExpressionDiscriminator.LOOP]
    # The actual expression value
    value: LoopValue


class LoopValue(TypedDict):
    """The actual expression value for a Viv loop."""
    # An expression that should evaluate to a value that is iterable in the runtime at hand.
    iterable: Expression
    # The local variable to which each member of the iterable is assigned on its respective iteration of the loop
    variable: LocalVariable
    # The body of the loop, structured as a list of expressions that will each be interpreted,
    # in order, on each iteration. These body expressions can reference the loop variable,
    # allowing for Viv code that acts on each member of an iterable.
    body: list[Expression]


class LocalVariable(TypedDict):
    """A Viv local variable."""
    # The name of the local variable
    name: VariableName
    # Whether the variable is marked as binding an entity (as opposed to a symbol)
    isEntityVariable: bool


# The name for a variable used in an assignment or a loop
VariableName: TypeAlias = str


class MembershipTest(SourceAnnotatedExpression, NegatableExpression):
    """A Viv membership test.

    This kind of expression takes two expressions as operands and evaluates to `True` if the evaluation
    of the first expression is a member of the evaluation of the second expression.
    """
    # Discriminator for a Viv membership test
    type: Literal[ExpressionDiscriminator.MEMBERSHIP_TEST]
    # The actual expression value
    value: MembershipTestValue


class MembershipTestValue(TypedDict):
    """The actual expression value for a Viv membership test."""
    # An expression whose evaluation will be used as the item operand in the membership test
    item: Expression
    # An expression whose evaluation will be used as the collection operand in the membership test
    collection: Expression


class MemoryCheck(SourceAnnotatedExpression, NegatableExpression):
    """A Viv memory check.

    This kind of expression takes two expressions as operands and evaluates to `True` if
    the evaluation of the first expression is a character, the second expression is an
    action, and the character has a memory of the action.
    """
    # Discriminator for a Viv memory check
    type: Literal[ExpressionDiscriminator.MEMORY_CHECK]
    # The actual expression value
    value: MemoryCheckValue


class MemoryCheckValue(TypedDict):
    """The actual expression value for a Viv memory check."""
    # An expression whose evaluation will be treated as the character operand in a memory check
    character: Expression
    # An expression whose evaluation will be treated as the action operand in a memory check
    action: Expression


class NullField(SourceAnnotatedExpression):
    """A Viv null value."""
    # Discriminator for a Viv null value
    type: Literal[ExpressionDiscriminator.NULL_TYPE]
    # Python's `None`, which serializes to JSON's `null`. In any Viv runtime, expressions of this
    # type evaluate to the null-like value in the language at hand.
    value: None


class ObjectField(SourceAnnotatedExpression):
    """A Viv object literal.

    Expressions of this type maps keys (string literals) to Viv expressions. Once evaluated,
    the result will map those same keys to the respective evaluations of the Viv expressions.
    """
    # Discriminator for a Viv object literal
    type: Literal[ExpressionDiscriminator.OBJECT]
    # The actual expression value
    value: dict[str, "Expression"]


class ActionSearch(SourceAnnotatedExpression):
    """A Viv action search.

    This kind of expression specifies a search over either a character's memories or the chronicle
    (all historical actions), which will return either that collection of actions, or a subset of
    them that match a specified target query.
    """
    # Discriminator for a Viv action search
    type: Literal[ExpressionDiscriminator.ACTION_SEARCH]
    # The actual expression value
    value: ActionSearchValue


class ActionSearchValue(TypedDict):
    """The actual expression value for a Viv action search."""
    # The name of the query that will be used for the search, if any, else `None`. If none
    # is provided, the search will simply retrieve all actions in the domain.
    queryName: QueryName | None
    # Precast bindings for the target query, as asserted in the search expression
    bindings: PrecastBindings
    # A declaration for how to construct a search domain for this action search
    searchDomain: SearchDomainDeclaration


class SearchDomainDeclaration(TypedDict):
    """A declaration for how to prepare a search domain for an action search or a sifting."""
    # The policy to use when preparing the search domain
    policy: SearchDomainPreparationPolicy
    # If the search is to run over a character's memories, this is an expression that should evaluate to
    # a reference to that character. If `None` is here, the query will be run over the chronicle (all
    # historical actions), though an error will be thrown if a query is specified and it uses memory
    # criteria (the `salience` and/or `associations` fields). A few notes: The compiler will ensure
    # that a value is only present here given an expression `policy`. If there is already a search
    # domain enclosing this one, the resulting domain will  be narrowed to the intersection of the
    # two domains (the existing one and the memories of the character specified by this expression).
    # If a character has forgotten an action, it will not be included in the constructed search domain.
    expression: Expression | None


class SearchDomainPreparationPolicy(StrEnum):
    """Enum containing the valid policies for preparing search domains for story sifting."""
    # Search in the full chronicle (all historical actions)
    CHRONICLE = "chronicle"
    # Search the memories of the character specified by the domain expression
    EXPRESSION = "expression"
    # Inherit the search domain passed into the evaluation context for the action search or sifting
    INHERIT = "inherit"


class Reaction(SourceAnnotatedExpression):
    """A Viv reaction expression.

    A reaction specifies an action that may be queued up for some time in the future,
    should an instance of the one at hand be performed.
    """
    # Discriminator for Viv reaction expressions
    type: Literal[ExpressionDiscriminator.REACTION]
    # The actual expression value
    value: ReactionValue


class ReactionValue(TypedDict):
    """A specification of the parameters defining a reaction."""
    # The name of the target construct (action or selector), i.e., the one queued up by this reaction
    targetName: ActionName | PlanName | SelectorName
    # The type of construct that the reaction will queue
    targetType: ReactionTargetConstructDiscriminator
    # Precast bindings for the target action, as asserted in the reaction declaration
    bindings: PrecastBindings
    # An expression that should evaluate to a boolean value indicating whether the reaction
    # will queue its target construct urgently. The evaluated value will be cast to a boolean,
    # so authors should be careful when using this parameter. Urgent actions and action selectors
    # receive the highest priority in action queues, while an urgent plan or plan selector will
    # be targeted (and potentially launched) upon being queued. If no expression is supplied,
    # the reaction will not be marked urgent.
    urgent: Expression | None
    # An expression that should evaluate to a numeric value specifying the priority of the queued
    # action. Within a given queue group (urgent or non-urgent), queued actions are targeted in
    # descending order of priority. The compiler will flag the use of this option in a reaction
    # queueing a plan or plan selector.
    priority: Expression | None
    # An expression that should evaluate to a location, that being the specific location
    # at which the reaction must be performed. The compiler will flag the use of this
    # option in a reaction queueing a plan or plan selector.
    location: list[SetPredicate] | None
    # A set of 0-3 temporal constraints constraining the time at which this reaction may be
    # performed. The compiler will flag the use of this option in a reaction queueing a plan
    # or plan selector.
    time: list[TemporalConstraint] | None
    # An object specifying logic around repeating the reaction should it succeed. (Viv already has
    # automatic retry logic for reactions that fail.)
    repeatLogic: ReactionRepeatLogic | None
    # A set of expressions such that, if all of them hold (i.e., evaluate to a truthy value),
    # the reaction will be dequeued.
    abandonmentConditions: list[Expression] | None


class ReactionRepeatLogic(TypedDict):
    """An object specifying logic around repeating the reaction after it *succeeds*.

    This logic is only considered when the construct queued by the reaction is successfully
    performed/executed. Note that Viv already has automatic retry logic for reactions that fail.
    """
    # A set of expressions such that, if all of them hold (i.e., evaluate to a truthy value),
    # copy of the reaction will be queued again for its initiator.
    conditions: list[Expression]
    # The maximum number of times this reaction may be repeated
    maxInstances: int


class PrecastBindings(TypedDict):
    """Precast bindings for the targeting of some construct, as in a reaction or a sifting expression."""
    # Whether the precast bindings are marked partial, meaning additional required role
    # slots will need to be cast. The compiler will verify that all required slots
    # appear to be precast, but we will also confirm this at runtime.
    partial: bool
    # A mapping from role name to an expression that will evaluate to precast bindings for that role
    roles: dict[RoleName, Expression]


class SetPredicate(TypedDict):
    """A set predicate that allows for certain values to appear in a set against which the operand will be tested."""
    # The operator associated with this predicate
    operator: SetPredicateOperator
    # The operand to use when testing this predicate
    operand: list[Expression]


# Enum containing the valid operators for the set predicates
SetPredicateOperator = Literal["none", "any", "all", "exactly"]


class Sifting(SourceAnnotatedExpression):
    """A Viv sifting.

    This kind of expression tests a sifting pattern against either the chronicle,
    meaning all historical actions, or a character's memories.
    """
    # Discriminator for a Viv sifting
    type: Literal[ExpressionDiscriminator.SIFTING]
    # The actual expression value
    value: SiftingValue


class SiftingValue(TypedDict):
    """The actual expression value for a Viv sifting."""
    # The name of the sifting pattern to run
    patternName: SiftingPatternName
    # Precast bindings for the target pattern, as asserted in the sifting expression
    bindings: PrecastBindings
    # A declaration for how to construct a search domain for this sifting
    searchDomain: SearchDomainDeclaration


class StringField(SourceAnnotatedExpression):
    """A Viv string literal."""
    # Discriminator for a Viv string literal
    type: Literal[ExpressionDiscriminator.STRING]
    # The string literal to which this expression will evaluate
    value: str


class TemplateStringField(SourceAnnotatedExpression):
    """A Viv templated string.

    This kind of expression is structured as an ordered list of string literals and string-producing
    expressions, the evaluations of which are concatenated to form the rendered string.
    """
    # Discriminator for a Viv templated string
    type: Literal[ExpressionDiscriminator.TEMPLATE_STRING]
    # The actual expression value
    value: list[str | Expression]


class TimeFrameStatement(TypedDict):
    """A temporal constraint specifying a range between two points in time."""
    # Discriminator for a time-frame temporal constraint
    type: Literal[TemporalStatementDiscriminator.TIME_FRAME]
    # The point in time at which the range opens. This is specified as a time delta (e.g., `5 days`) that
    # the host application can resolve at runtime into a point in time (by anchoring it relative to a given
    # simulation timestamp). If no delta is specified here, the range is open on this end.
    open: TimeDelta | None
    # The point in time at which the range closes. This is specified as a time delta (e.g., `5 days`) that
    # the host application can resolve at runtime into a point in time (by anchoring it relative to a given
    # simulation timestamp). If no delta is specified here, the range is open on this end.
    close: TimeDelta | None
    # For reactions only: whether to anchor the time-frame constraint in the timestamp of the action that directly
    # triggered this reaction -- meaning the action whose definition included the reaction declaration -- as opposed
    # to the current simulation timestamp. This distinction only matters for cases where a reaction is triggered
    # because a character has learned about an action after the fact. In such cases, we need to know whether a
    # time-frame constraint like "between 1 year and 3 years" holds relative to the timestamp of the original
    # action or relative to the time at which the character learned about the original action.
    useActionTimestamp: bool


class TimeDelta(TypedDict):
    """A time period (e.g., "2 weeks") that can be used as a delta to resolve a point
    in time relative to some anchor time (e.g., the current simulation time).
    """
    # The number of time units -- e.g., `2` in `2 weeks`
    amount: float
    # The unit of time -- e.g., `weeks` in `2 weeks`
    unit: TimeFrameTimeUnit


class TimeOfDayStatement(TypedDict):
    """A temporal constraint specifying a range between two times of day."""
    # Discriminator for a time-of-day temporal constraint
    type: Literal[TemporalStatementDiscriminator.TIME_OF_DAY]
    # The time of day that opens the range. If no time of day is specified here, the range
    # is open on this end. Note that the host application is tasked with determining
    # whether a given time of day has passed.
    open: TimeOfDayDeclaration | None
    # The time of day that closes the range. If no time of day is specified here, the range
    # is open on this end. Note that the host application is tasked with determining
    # whether a given time of day has passed.
    close: TimeOfDayDeclaration | None


class TimeOfDayDeclaration(TypedDict):
    """A specified time of day."""
    # The hour of day (in `[0, 23]`)
    hour: int
    # The minute of the hour of day (in `[0, 59]`)
    minute: int


# Enum containing the valid time units for time-frame statements
TimeFrameTimeUnit = Literal["minutes", "hours", "days", "weeks", "months", "years"]


class TemporalStatementDiscriminator(StrEnum):
    """Enum containing discriminators for the possible temporal constraints."""
    # Discriminator for a time-frame temporal constraint
    TIME_FRAME = "timeFrame"
    # Discriminator for a time-of-day temporal constraint
    TIME_OF_DAY = "timeOfDay"


# A temporal constraint specifying a range between either points in time or times of day
TemporalConstraint = TimeFrameStatement | TimeOfDayStatement


class TropeFit(SourceAnnotatedExpression, NegatableExpression):
    """A Viv "trope fit" expression.

    This kind of expression evaluates to `True` if the trope holds with the given arguments.
    """
    # Discriminator for a Viv trope-fit expression
    type: Literal[ExpressionDiscriminator.TROPE_FIT]
    # The actual expression value
    value: TropeFitValue


class TropeFitValue(TypedDict):
    """The actual expression value for a Viv trope-fit expression."""
    # The name of the trope that will be used for the test
    tropeName: TropeName
    # Precast bindings for the target trope, as asserted in the trope-fit expression
    bindings: PrecastBindings



"""Mixin class containing visitor methods associated with Viv statements."""

__all__ = ["VisitorMixinStatements"]

from typing import Any

from arpeggio import NonTerminal, PTNodeVisitor

from viv_compiler import external_types, sentinels, utils


class VisitorMixinStatements(PTNodeVisitor):
    """A visitor mixin for Viv statements."""

    @staticmethod
    def visit_statements(_, children: list[external_types.Expression]) -> list[external_types.Expression]:
        """Visit a <statements> node."""
        return children

    @staticmethod
    def visit_scoped_statements(_, children: list[external_types.Expression]) -> list[external_types.Expression]:
        """Visit a <scoped_statements> node."""
        return children

    @staticmethod
    def visit_statement(_, children: list[external_types.Expression]) -> external_types.Expression:
        """Visit a <statement> node."""
        return children[0]

    @staticmethod
    def visit_conditional(node: NonTerminal, children: list[Any]) -> external_types.Conditional:
        """Visit a <conditional> node."""
        if len(children) > 1:
            branches, alternative = children
        else:
            branches, alternative = children[0], None
        conditional = external_types.Conditional(
            type=external_types.ExpressionDiscriminator.CONDITIONAL,
            value=external_types.ConditionalValue(branches=branches, alternative=alternative),
            source=utils.derive_source_annotations(node=node)
        )
        return conditional

    @staticmethod
    def visit_conditional_branches(
        _, children: list[external_types.ConditionalBranch]
    ) -> list[external_types.ConditionalBranch]:
        """Visit a <conditional_branches> node."""
        return children

    @staticmethod
    def visit_conditional_branch(_, children: list[Any]) -> external_types.ConditionalBranch:
        """Visit a <conditional_branch> node."""
        condition, consequent = children
        conditional_branch = external_types.ConditionalBranch(condition=condition, consequent=consequent)
        return conditional_branch

    @staticmethod
    def visit_condition(_, children: list[external_types.Expression]) -> external_types.Expression:
        """Visit a <condition> node."""
        return children[0]

    @staticmethod
    def visit_consequent(_, children: list[list[external_types.Expression]]) -> list[external_types.Expression]:
        """Visit a <consequent> node."""
        return children[0]

    @staticmethod
    def visit_alternative(_, children: list[list[external_types.Expression]]) -> list[external_types.Expression]:
        """Visit an <alternative> node."""
        return children[0]

    @staticmethod
    def visit_loop(node: NonTerminal, children: list[Any]) -> external_types.Loop:
        """Visit a <loop> node."""
        iterable_reference, loop_variable, loop_body = children
        loop = external_types.Loop(
            type=external_types.ExpressionDiscriminator.LOOP,
            value=external_types.LoopValue(iterable=iterable_reference, variable=loop_variable, body=loop_body),
            source=utils.derive_source_annotations(node=node)
        )
        return loop

    @staticmethod
    def visit_local_variable(_, children: list[Any]) -> external_types.LocalVariable:
        """Visit a <local_variable> node."""
        _sigil, binding_type, name = children
        local_variable = external_types.LocalVariable(
            name=name,
            isEntityVariable=binding_type is sentinels.ENTITY_BINDING_TYPE
        )
        return local_variable



"""Module entry point for the `python -m viv_compiler` usage pattern."""

from .cli import main


# When the package is executed as a module, we simply invoke the Viv compiler CLI (`vivc`)
if __name__ == "__main__":
    main()



"""Mixin class containing visitor methods associated with precast bindings."""

__all__ = ["VisitorMixinBindings"]

from typing import Any

from arpeggio import NonTerminal, PTNodeVisitor

from viv_compiler import config, errors, external_types, internal_types, utils


class VisitorMixinBindings(PTNodeVisitor):
    """A visitor mixin for Viv precast bindings."""

    @staticmethod
    def visit_bindings(_, children: list[Any]) -> dict[str, external_types.PrecastBindings]:
        """Visit a <bindings> node."""
        return {"bindings": children[0]}

    @staticmethod
    def visit_bindings_none(_, __) -> external_types.PrecastBindings:
        """Visit a <bindings_none> node."""
        precast_bindings = external_types.PrecastBindings(partial=True, roles={})
        return precast_bindings

    @staticmethod
    def visit_bindings_partial(
        node: NonTerminal,
        children: list[tuple[external_types.RoleName, external_types.Expression]]
    ) -> external_types.PrecastBindings:
        """Visit a <bindings_partial> node."""
        precast_bindings = external_types.PrecastBindings(
            partial=True,
            roles=VisitorMixinBindings._collect_bindings(node=node, bindings=children)
        )
        return precast_bindings

    @staticmethod
    def visit_bindings_complete(node: NonTerminal, children: list[Any]) -> external_types.PrecastBindings:
        """Visit a <bindings_complete> node."""
        precast_bindings = external_types.PrecastBindings(
            partial=False,
            roles=VisitorMixinBindings._collect_bindings(node=node, bindings=children)
        )
        return precast_bindings

    @staticmethod
    def visit_bindings_sugared(_, children: list[Any]) -> dict[str, external_types.PrecastBindings]:
        """Visit a <bindings_sugared> node."""
        return {"bindings": children[0]}

    @staticmethod
    def visit_bindings_sugared_none(_, __) -> external_types.PrecastBindings:
        """Visit a <bindings_sugared_none> node."""
        return external_types.PrecastBindings(partial=True, roles={})

    @staticmethod
    def visit_bindings_sugared_partial(node: NonTerminal, children: list[Any]) -> external_types.PrecastBindings:
        """Visit a <bindings_sugared_partial> node."""
        precast_bindings = external_types.PrecastBindings(
            partial=True,
            roles=VisitorMixinBindings._collect_bindings(node=node, bindings=children[0])
        )
        return precast_bindings

    @staticmethod
    def visit_bindings_sugared_complete(node: NonTerminal, children: list[Any]) -> external_types.PrecastBindings:
        """Visit a <bindings_sugared_complete> node."""
        precast_bindings = external_types.PrecastBindings(
            partial=False,
            roles=VisitorMixinBindings._collect_bindings(node=node, bindings=children[0])
        )
        return precast_bindings

    @staticmethod
    def visit_bindings_sugared_actual_bindings(_, children: list[Any]) -> list[dict[str, Any]]:
        """Visit a <bindings_sugared_actual_bindings> node."""
        return children

    @staticmethod
    def _collect_bindings(
        node: NonTerminal,
        bindings: list[tuple[external_types.RoleName | None, external_types.Expression]]
    ) -> dict[external_types.RoleName, external_types.Expression]:
        """Prepare the 'roles' field of a precast bindings."""
        roles: dict[external_types.RoleName, external_types.Expression] = {}
        for i, (role_name, expression) in enumerate(bindings):
            if role_name is None:  # Positional binding
                role_name = f"{config.POSITIONAL_BINDING_TRANSIENT_PREFIX}{i}"
            if role_name in roles:
                error_message = (
                    f"Duplicate role detected in precast bindings (use comma separation to "
                    f"precast multiple slots): '{role_name}'"
                )
                raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
            roles[role_name] = expression
        return roles

    @staticmethod
    def visit_binding(
        _, children: list[internal_types.IntermediateRoleReference, external_types.Expression]
    ) -> tuple[external_types.RoleName, external_types.Expression]:
        """Visit a <binding> node."""
        role_reference, expression = children
        role_reference: internal_types.IntermediateRoleReference
        expression: external_types.Expression
        binding = (role_reference["_name"], expression)
        return binding

    @staticmethod
    def visit_positional_binding(
        _, children: list[external_types.Expression]
    ) -> tuple[None, external_types.Expression]:
        """Visit a <positional_binding> node."""
        positional_binding = (None, children[0])
        return positional_binding



"""Mixin class containing visitor methods associated with sifting-pattern definitions."""

__all__ = ["VisitorMixinSiftingPatterns"]

from typing import Any

from arpeggio import NonTerminal, PTNodeVisitor

from viv_compiler import errors, external_types, internal_types, utils, validation


class VisitorMixinSiftingPatterns(PTNodeVisitor):
    """A visitor mixin for Viv sifting patterns."""

    @staticmethod
    def visit_sifting_pattern(
        node: NonTerminal, children: list[Any]
    ) -> internal_types.IntermediateSiftingPatternDefinition:
        """Visit a <sifting_pattern> node."""
        # Collect all role definitions, which may be distributed across the 'roles' and 'actions' sections
        sifting_pattern_name, sifting_pattern_body = children
        all_role_definitions = []
        if "roles" in sifting_pattern_body:
            all_role_definitions += sifting_pattern_body["roles"]
        if "actions" in sifting_pattern_body:
            all_role_definitions += sifting_pattern_body["actions"]
        else:
            error_message = f"Sifting pattern '{sifting_pattern_name}' is missing 'actions' field (required)"
            raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
        validation.prevalidate_role_names(
            construct_type=external_types.ConstructDiscriminator.SIFTING_PATTERN,
            construct_name=sifting_pattern_name,
            role_definitions=all_role_definitions,
            source=utils.derive_source_annotations(node=node)
        )
        # Populate the fields that were authored
        accumulated_fields = {
            "name": sifting_pattern_name,
            "roles": {role["name"]: role for role in all_role_definitions}
        }
        for field_name, field_value in sifting_pattern_body.items():
            if field_name == "roles":
                continue  # Already handled just above
            if field_name == "actions":
                accumulated_fields["actions"] = [action['name'] for action in field_value]
            else:
                accumulated_fields[field_name] = field_value
        # Package up the intermediate definition
        intermediate_sifting_pattern_definition = internal_types.IntermediateSiftingPatternDefinition(
            type=external_types.ConstructDiscriminator.SIFTING_PATTERN,
            name=sifting_pattern_name,
            roles=accumulated_fields["roles"],
            actions=accumulated_fields["actions"],
            _conditions_raw=accumulated_fields.get("_conditions_raw", []),
        )
        return intermediate_sifting_pattern_definition

    @staticmethod
    def visit_sifting_pattern_header(
        _, children: list[external_types.SiftingPatternName]
    ) -> external_types.SiftingPatternName:
        """Visit a <sifting_pattern_header> node."""
        return children[0]

    @staticmethod
    def visit_sifting_pattern_body(_, children: list[Any]) -> dict[str, Any]:
        """Visit a <sifting_pattern_body> node."""
        accumulated_fields = {}
        for child in children:
            accumulated_fields.update(child)
        return accumulated_fields

    @staticmethod
    def visit_sifting_pattern_roles(
        _, children: list[external_types.RoleDefinition]
    ) -> dict[str, list[external_types.RoleDefinition]]:
        """Visit a <sifting_pattern_roles> node."""
        return {"roles": children}

    @staticmethod
    def visit_sifting_pattern_actions(
        _, children: list[external_types.RoleDefinition]
    ) -> dict[str, list[external_types.RoleDefinition]]:
        """Visit a <sifting_pattern_actions> node."""
        return {"actions": children}

    @staticmethod
    def visit_sifting_pattern_action(_, children: list[Any]) -> external_types.RoleDefinition:
        """Visit a <sifting_pattern_action> node."""
        # The 'actions' field in a sifting pattern is just syntactic sugar for defining
        # an action role that will also be exposed as one of the sifted actions.
        _sigil, action_name, casting_pool_directive = children
        role_definition = external_types.RoleDefinition(
            name=action_name,
            entityType=external_types.RoleEntityType.ACTION,
            min=1,
            max=1,
            parent=None,
            children=[],
            pool=casting_pool_directive["pool"],
            chance=None,
            mean=None,
            sd=None,
            participationMode=None,
            anywhere=True,
            precast=False,
            spawn=False,
            spawnFunction=None,
            renames=None
        )
        return role_definition

    @staticmethod
    def visit_sifting_pattern_conditions(
        _, children: list[list[external_types.Expression]]
    ) -> dict[str, list[external_types.Expression]]:
        """Visit a <sifting_pattern_conditions> node."""
        return {"_conditions_raw": children[0]}



"""Internal functionality for validating selector definitions."""

__all__ = ["validate_action_selector_definitions", "validate_plan_selector_definitions"]

from viv_compiler import errors, external_types, utils
from ._common import validate_common_concerns
from ._utils import detect_cycle_in_graph, is_non_numeric_expression


def validate_action_selector_definitions(content_bundle: external_types.ContentBundle) -> None:
    """Validate the action-selector definitions in the given compiled content bundle.

    Args:
        content_bundle: A compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.
    """
    for action_selector_definition in content_bundle["actionSelectors"].values():
        # Validate general issues that pertain to all construct types
        validate_common_concerns(construct_definition=action_selector_definition, content_bundle=content_bundle)
        # Validate the initiator bindings for all selector candidate
        _validate_action_selector_candidate_initiator_bindings(
            action_selector_definition=action_selector_definition,
            content_bundle=content_bundle
        )
        # Validate issues that pertain to both selector types
        _validate_selector_common_concerns(
            selector_definition=action_selector_definition,
            selector_type=action_selector_definition['type'],
            content_bundle=content_bundle
        )


def _validate_action_selector_candidate_initiator_bindings(
    action_selector_definition: external_types.ActionSelectorDefinition,
    content_bundle: external_types.ContentBundle
) -> None:
    """Ensure that the given selector action-selector definition precasts its initiator
    in the respective initiator roles of its candidates.

    Args:
        action_selector_definition: The action-selector definition to validate.
        content_bundle: The compiled content bundle containing the selector definition.

    Returns:
        None. Only returns if no issue was detected downstream.
    """
    for candidate in action_selector_definition['candidates']:
        # Retrieve candidate definition
        if candidate['isSelector']:
            candidate_definition = content_bundle['actionSelectors'][candidate['name']]
        else:
            candidate_definition = content_bundle['actions'][candidate['name']]
        # Ensure that the initiator role of the candidate is precast
        if candidate_definition['initiator'] not in candidate['bindings']['roles']:
            raise errors.VivCompileError(
                f"Action selector '{action_selector_definition['name']}' fails to precast the initiator role "
                f"'{candidate_definition['initiator']}' for candidate '{candidate_definition['name']}'"
            )
        # Furthermore, ensure that the selector initiator is precast in the initiator role of the candidate,
        # which we  detect as an entity reference anchored in that role, with no path and no other properties.
        initiator_candidate_expression = candidate['bindings']['roles'][candidate_definition['initiator']]
        if not (
            initiator_candidate_expression['type'] == external_types.ExpressionDiscriminator.ENTITY_REFERENCE and
            initiator_candidate_expression['value']['anchor'] == action_selector_definition['initiator'] and
            not initiator_candidate_expression['value']['path'] and
            not initiator_candidate_expression['value']['local'] and
            not initiator_candidate_expression['value']['group']
        ):
            raise errors.VivCompileError(
                msg=(
                    f"Action selector '{action_selector_definition['name']}' fails to directly precast "
                    f"its initiator role '{action_selector_definition['initiator']}' in the corresponding "
                    f"initiator role '{candidate_definition['initiator']}' for candidate "
                    f"'{candidate_definition['name']}' (the initiator roles must line up directly): "
                    f"{initiator_candidate_expression['source']['code']}"
                ),
                source=initiator_candidate_expression['source'],
            )


def validate_plan_selector_definitions(content_bundle: external_types.ContentBundle) -> None:
    """Validate the plan-selector definitions in the given compiled content bundle.

    Args:
        content_bundle: A compiled content bundle.

    Returns:
        None. Only returns if no issue was detected downstream.
    """
    for plan_selector_definition in content_bundle["planSelectors"].values():
        # Validate general issues that pertain to all construct types
        validate_common_concerns(construct_definition=plan_selector_definition, content_bundle=content_bundle)
        # Ensure that selector candidates are all defined
        _detect_undefined_plan_selector_candidate(
            plan_selector_definition=plan_selector_definition,
            content_bundle=content_bundle
        )
        # Validate issues that pertain to both selector types
        _validate_selector_common_concerns(
            selector_definition=plan_selector_definition,
            selector_type=plan_selector_definition['type'],
            content_bundle=content_bundle
        )
    # Detect any cycle among plan selectors
    _detect_plan_selector_candidate_cycle(content_bundle=content_bundle)


def _detect_undefined_plan_selector_candidate(
    plan_selector_definition: external_types.PlanSelectorDefinition,
    content_bundle: external_types.ContentBundle
) -> None:
    """Ensure that all the given plan selector's candidates are defined plans and/or plan selectors.

    Args:
        plan_selector_definition: The plan-selector definition to validate.
        content_bundle: The compiled content bundle containing the plan-selector definition.

    Returns:
        None.

    Raises:
        VivCompileError: The plan selector has an undefined candidate.
    """
    for candidate in plan_selector_definition["candidates"]:
        if candidate["isSelector"]:
            if not any(selector['name'] == candidate['name'] for selector in content_bundle["planSelectors"].values()):
                raise errors.VivCompileError(
                    f"Plan selector '{plan_selector_definition['name']}' targets undefined "
                    f"plan selector: '{candidate['name']}'"
                )
        elif not any(plan['name'] == candidate['name'] for plan in content_bundle["plans"].values()):
            raise errors.VivCompileError(
                f"Plan selector '{plan_selector_definition['name']}' targets undefined plan: '{candidate['name']}'"
            )


def _detect_plan_selector_candidate_cycle(content_bundle: external_types.ContentBundle) -> None:
    """Throws an error if there is a cycle among the plan-selector definitions in the given content bundle.

    A selector can yield a cycle when it targets itself, or when it targets another selector
    that directly or indirectly ends up targeting the original selector.

    Args:
        content_bundle: A compiled content bundle.

    Returns:
        None. Only returns if there's no cycle.
    """
    # Build adjacency graph
    adjacency_graph = {plan_selector_name: set() for plan_selector_name in content_bundle["planSelectors"]}
    for selector_definition in content_bundle["planSelectors"].values():
        for candidate in selector_definition["candidates"]:
            if candidate["isSelector"]:
                if candidate["name"] == selector_definition["name"]:
                    raise errors.VivCompileError(
                        f"Plan selector '{selector_definition['name']}' directly targets itself "
                        f"(prohibited because it would yield an infinite loop during targeting)"
                    )
                adjacency_graph[selector_definition['name']].add(candidate["name"])
    # Attempt to select an indirect cycle
    cycle = detect_cycle_in_graph(adjacency_graph=adjacency_graph)
    if cycle:
        cycle_str = "'" + "' > '".join(cycle) + "'"
        raise errors.VivCompileError(
            f"Cycle detected among plan selectors: {cycle_str} "
            f"(prohibited because such cycles make selector targeting impossible to resolve)"
        )


def _validate_selector_common_concerns(
    selector_definition: external_types.SelectorDefinitionBase,
    selector_type: external_types.ConstructDiscriminator,
    content_bundle: external_types.ContentBundle
) -> None:
    """Validate the given selector definition (action selector or plan selector) along several concerns
    that are common to both selector types.

    Args:
        selector_definition: The selector definition to validate. This is typed as the selector base
            class due to some weirdness in the type checker, which doesn't understand that the
            actual selector classes inheriting from this base class have all its properties.
        content_bundle: The compiled content bundle containing the selector definition.

    Returns:
        None. Only returns if no issue was detected downstream.
    """
    _validate_selector_candidate_names(selector_definition=selector_definition, selector_type=selector_type)
    _validate_selector_candidate_weights(selector_definition=selector_definition, selector_type=selector_type)
    _validate_selector_precast_bindings(
        selector_definition=selector_definition,
        selector_type=selector_type,
        content_bundle=content_bundle
    )


def _validate_selector_candidate_names(
    selector_definition: external_types.SelectorDefinitionBase,
    selector_type: external_types.ConstructDiscriminator,
) -> None:
    """Validate the candidate names in the given selector definition.

    Args:
        selector_definition: The selector definition to validate.
        selector_type: Construct discriminator for the selector.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: A duplicate candidate name was encountered.
    """
    for i, candidate in enumerate(selector_definition['candidates']):
        for j, other_candidate in enumerate(selector_definition['candidates']):
            if i == j:
                continue
            if candidate['name'] == other_candidate['name']:
                if candidate['isSelector'] == other_candidate['isSelector']:
                    raise errors.VivCompileError(
                        f"{utils.CONSTRUCT_LABEL[selector_type]} "
                        f"'{selector_definition['name']}' has duplicate candidate name: "
                        f"'{'selector ' if candidate['isSelector'] else ''}{candidate['name']}'"
                    )


def _validate_selector_candidate_weights(
    selector_definition: external_types.SelectorDefinitionBase,
    selector_type: external_types.ConstructDiscriminator
) -> None:
    """Validate the use of candidate weights in the given selector definition.

    Args:
        selector_definition: The selector definition to validate.
        selector_type: Construct discriminator for the selector.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The selector is missing weights, has bad weights, or uses weights when it shouldn't.
    """
    for candidate in selector_definition['candidates']:
        # If the selector uses the weighted-random policy, ensure that the candidate has a valid associated weight
        if selector_definition['policy'] == external_types.SelectorPolicy.WEIGHTED:
            if not candidate['weight']:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[selector_type]} '{selector_definition['name']}' "
                    f"uses weighted-random targeting policy, but candidate '{candidate['name']}' "
                    f"has not been attributed a weight"
                )
            elif is_non_numeric_expression(expression=candidate['weight']):
                raise errors.VivCompileError(
                    msg=(
                        f"{utils.CONSTRUCT_LABEL[selector_type]} '{selector_definition['name']}' "
                        f"uses an invalid expression type to specify the weight for candidate "
                        f"'{candidate['name']}' (value will be non-numeric): "
                        f"{candidate['weight']['source']['code']}"
                    ),
                    source=candidate['weight']['source'],
                )
            elif candidate['weight']['type'] in (
                    external_types.ExpressionDiscriminator.FLOAT,
                    external_types.ExpressionDiscriminator.INT
            ):
                if candidate['weight']['value'] < 0:
                    raise errors.VivCompileError(
                        msg=(
                            f"{utils.CONSTRUCT_LABEL[selector_type]} "
                            f"'{selector_definition['name']}' has negative weight for candidate "
                            f"'{candidate['name']}' (weights must be zero or greater): "
                            f"{candidate['weight']['source']['code']}"
                        ),
                        source=candidate['weight']['source'],
                    )
        elif candidate['weight']:
            # Otherwise, ensure that the candidate does not have an associated weight
            raise errors.VivCompileError(
                f"{utils.CONSTRUCT_LABEL[selector_type]} '{selector_definition['name']}' "
                f"does not use the weighted-random targeting policy, but candidate "
                f"'{candidate['name']}' has still been attributed a weight"
            )


def _validate_selector_precast_bindings(
    selector_definition: external_types.SelectorDefinitionBase,
    selector_type: external_types.ConstructDiscriminator,
    content_bundle: external_types.ContentBundle
) -> None:
    """Validate the precast bindings passed onto the respective candidates in the given selector definition.

    Args:
        selector_definition: The selector definition to validate.
        selector_type: Construct discriminator for the selector.
        content_bundle: The compiled content bundle containing the selector definition.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: An issue was detected in the precast bindings for one of the action-selector candidates.
    """
    for candidate in selector_definition['candidates']:
        if selector_type == external_types.ConstructDiscriminator.ACTION_SELECTOR:
            if candidate['isSelector']:
                candidate_definition = content_bundle['actionSelectors'][candidate['name']]
            else:
                candidate_definition = content_bundle['actions'][candidate['name']]
        elif selector_type == external_types.ConstructDiscriminator.PLAN_SELECTOR:
            if candidate['isSelector']:
                candidate_definition = content_bundle['planSelectors'][candidate['name']]
            else:
                candidate_definition = content_bundle['plans'][candidate['name']]
        else:
            raise errors.VivCompileError(
                f"Unexpected construct type for selector: '{selector_type}'"
            )
        # Detect missing required roles
        if not candidate['bindings']['partial']:
            for role_definition in candidate_definition["roles"].values():
                if role_definition['min'] > 0:
                    if role_definition['name'] not in candidate['bindings']['roles']:
                        raise errors.VivCompileError(
                            f"{utils.CONSTRUCT_LABEL[selector_type]} "
                            f"'{selector_definition['name']}' fails to cast required role "
                            f"'{role_definition['name']}' in bindings for candidate '{candidate['name']}' "
                            f"(use 'partial' to pass a partial cast)"
                        )
        # Detect missing precast roles
        for role_definition in candidate_definition['roles'].values():
            if role_definition['precast']:
                if role_definition['name'] not in candidate['bindings']['roles']:
                    raise errors.VivCompileError(
                        f"{utils.CONSTRUCT_LABEL[selector_type]} "
                        f"'{selector_definition['name']}' fails to precast a 'precast' role "
                        f"for candidate '{candidate['name']}': '{role_definition['name']}'"
                    )
        # Detect extra roles
        for role_name in candidate['bindings']['roles']:
            if role_name not in candidate_definition['roles']:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[selector_type]} "
                    f"'{selector_definition['name']}' precasts role '{role_name}' in "
                    f"bindings for candidate '{candidate['name']}', but candidate has no role by that name"
                )



"""Utility functions used across the compiler modules."""

from pathlib import Path
from typing import Any, Callable, Final

from arpeggio import NonTerminal, SemanticActionResults

from viv_compiler import config, external_types, internal_types


# A mapping from construct discriminator to a string suitable for identifying the construct type in an error message
CONSTRUCT_LABEL: Final = {
    external_types.ConstructDiscriminator.ACTION: "Action",
    external_types.ConstructDiscriminator.ACTION_SELECTOR: "Action selector",
    external_types.ConstructDiscriminator.PLAN: "Plan",
    external_types.ConstructDiscriminator.PLAN_SELECTOR: "Plan selector",
    external_types.ConstructDiscriminator.QUERY: "Query",
    external_types.ConstructDiscriminator.SIFTING_PATTERN: "Sifting pattern",
    external_types.ConstructDiscriminator.TROPE: "Trope",
}

# Absolute path to the Viv source file whose parse tree is currently being traversed by the Visitor
SOURCE_FILE_PATH: Path | None = None

# The contents of the Viv source file whose parse tree is currently being traversed by the Visitor
SOURCE_FILE_CONTENTS: str | None = None

# Handle on the Arpeggio ParserPEG method for converting a position index from the source code into line
# and column number. This comes from the parser, which uses transient data collecting during parsing.
POSITION_TO_LINE_AND_COLUMN: Callable | None = None


def register_source_file(
    source_file_path: Path,
    source_file_contents: str,
    position_to_line_and_column: Callable
):
    """Registers metadata about the Viv source file whose parse tree is currently being traversed.

    Args:
        source_file_path: Absolute path to the Viv source file whose parse tree is currently
            being traversed by the Visitor.
        source_file_contents: The contents of the Viv source file whose parse tree is currently
            being traversed by the Visitor
        position_to_line_and_column: Handle on the Arpeggio ParserPEG method for converting
            a position index from the source code into line and column number. This comes from
            the parser, which uses transient data collecting during parsing.

    Returns:
        Nothing. The metadata is stored in the global constants above.
    """
    global SOURCE_FILE_PATH, SOURCE_FILE_CONTENTS, POSITION_TO_LINE_AND_COLUMN
    SOURCE_FILE_PATH = source_file_path
    SOURCE_FILE_CONTENTS = source_file_contents
    POSITION_TO_LINE_AND_COLUMN = position_to_line_and_column


def derive_source_annotations(node: NonTerminal) -> external_types.SourceAnnotations:
    """Return source annotations derived for the given expression AST node.

    Args:
        node: The AST node for which source annotations will be derived.

    Returns:
        Source annotations derived for the given expression AST node.

    Raises:
        RuntimeError: The source file has not yet been registered. This is only caused by a
            fundamental compiler issue, which should never occur for an end user.
    """
    if not (SOURCE_FILE_PATH and SOURCE_FILE_CONTENTS and POSITION_TO_LINE_AND_COLUMN):
        raise RuntimeError("Source file must be registered prior to deriving source annotations")
    start_line_number, start_column_number = POSITION_TO_LINE_AND_COLUMN(node.position)
    end_line_number, end_column_number = POSITION_TO_LINE_AND_COLUMN(node.position_end)
    code = SOURCE_FILE_CONTENTS[node.position:node.position_end]
    source_annotations = external_types.SourceAnnotations(
        filePath=str(SOURCE_FILE_PATH),
        line=start_line_number,
        column=start_column_number,
        endLine=end_line_number,
        endColumn=end_column_number,
        code=code.rstrip()  # Strip trailing whitespace
    )
    return source_annotations


def sanitize_ast(ast: Any) -> Any:
    """Returns a deep copy of the given AST with all (nested) Arpeggio containers replaced with plain lists.

    Args:
        ast: An AST produced by our visitor.

    Returns:
        A deep copy of the given AST with all (nested) Arpeggio containers replaced with plain lists.
    """
    if isinstance(ast, SemanticActionResults):
        return [sanitize_ast(value) for value in ast]
    if isinstance(ast, dict):
        return {key: sanitize_ast(value) for key, value in ast.items()}
    if isinstance(ast, list):
        return [sanitize_ast(value) for value in ast]
    if isinstance(ast, tuple):  # We'll be serializing to JSON soon enough, so let's convert to a list right now
        return [sanitize_ast(value) for value in ast]
    return ast


def get_all_construct_definitions(
    combined_ast: internal_types.CombinedAST
) -> list[internal_types.IntermediateConstructDefinition]:
    """Returns a list containing all intermediate construct definitions in the given combined AST.

    Args:
        combined_ast: A combined AST at any point in the postprocessing pipeline.

    Returns:
        A list containing all construct definitions in the given combined AST.
    """
    all_construct_definitions = (
        combined_ast['actions'] +
        combined_ast['actionSelectors'] +
        combined_ast['plans'] +
        combined_ast['planSelectors'] +
        combined_ast['queries'] +
        combined_ast['siftingPatterns'] +
        combined_ast['tropes']
    )
    return all_construct_definitions


def get_all_referenced_roles(ast_chunk: Any) -> list[external_types.RoleName]:
    """Return a deduplicated list of all roles referenced in the given AST chunk.

    Note: Order must be preserved here.

    Args:
        ast_chunk: The full or partial AST to search for role references.

    Returns:
        A list containing the names of all roles referenced in the given AST chunk.
    """
    reference_discriminators = (
        external_types.ExpressionDiscriminator.ENTITY_REFERENCE,
        external_types.ExpressionDiscriminator.SYMBOL_REFERENCE
    )
    roles_referenced_so_far: list[external_types.RoleName] = []
    if isinstance(ast_chunk, list):
        for element in ast_chunk:
            roles_referenced_so_far.extend(get_all_referenced_roles(ast_chunk=element))
    elif isinstance(ast_chunk, dict):
        ast_chunk_type = ast_chunk.get('type')
        if ast_chunk_type in reference_discriminators:
            if not ast_chunk['value']['local']:
                if ast_chunk['value']['anchor'] != config.ACTION_SELF_REFERENCE_ROLE_NAME:
                    roles_referenced_so_far.append(ast_chunk['value']['anchor'])
        for value in ast_chunk.values():
            roles_referenced_so_far.extend(get_all_referenced_roles(ast_chunk=value))
    return sorted(set(roles_referenced_so_far))


def get_all_expressions_of_type(expression_type: external_types.ExpressionDiscriminator, ast_chunk: Any) -> list[Any]:
    """Return a list containing values for all expressions of the given type that are nested in the given AST chunk.

    Args:
        expression_type: String indicating the type of Viv expression to search for.
        ast_chunk: The AST chunk to search for expressions of the given type.

    Returns:
        A list containing all the Viv expressions of the given type in the given AST chunk.
    """
    expressions = []
    if isinstance(ast_chunk, list):
        for element in ast_chunk:
            expressions.extend(get_all_expressions_of_type(expression_type=expression_type, ast_chunk=element))
    elif isinstance(ast_chunk, dict):
        if 'type' in ast_chunk and ast_chunk['type'] == expression_type:
            expression_of_type = ast_chunk['value']
            expressions.append(expression_of_type)
        for _key, value in ast_chunk.items():  # No `else` here, because we want to recurse into the expression value
            expressions.extend(get_all_expressions_of_type(expression_type=expression_type, ast_chunk=value))
    return expressions


def get_all_full_expressions_of_type(
    expression_type: external_types.ExpressionDiscriminator,
    ast_chunk: Any
) -> list[Any]:
    """Return a list containing full expression dicts for all expressions of the given type in the given AST chunk.

    Unlike `get_all_expressions_of_type`, which returns only the inner `value` field, this function returns the
    complete expression dict (including `type`, `value`, and `source`). This is useful when source annotations
    are needed for error reporting.

    Args:
        expression_type: String indicating the type of Viv expression to search for.
        ast_chunk: The AST chunk to search for expressions of the given type.

    Returns:
        A list containing full expression dicts for all matching expressions in the given AST chunk.
    """
    expressions = []
    if isinstance(ast_chunk, list):
        for element in ast_chunk:
            expressions.extend(get_all_full_expressions_of_type(expression_type=expression_type, ast_chunk=element))
    elif isinstance(ast_chunk, dict):
        if 'type' in ast_chunk and ast_chunk['type'] == expression_type:
            expressions.append(ast_chunk)
        for _key, value in ast_chunk.items():  # No `else` here, because we want to recurse into the expression value
            expressions.extend(get_all_full_expressions_of_type(expression_type=expression_type, ast_chunk=value))
    return expressions


def is_initiator_role(role_definition: external_types.RoleDefinition) -> bool:
    """Return whether the given role is an initiator role.

    Args:
        role_definition: Definition for the role in question.

    Returns:
        Whether the given role is an initiator role.
    """
    return role_definition["participationMode"] == external_types.RoleParticipationMode.INITIATOR
